# type: ignore
def appproximate_to_tradeoff(
    param: tuple[f64, f64]
) -> tuple[Callable[[RBig], RBig], RBig]:
    epsilon, delta = param

    exp_eps = epsilon.with_rounding(Down).exp()  # `\label{exp-eps}`
    exp_eps = RBig.try_from(exp_eps)

    exp_neg_eps = (-epsilon).with_rounding(Up).exp()  # `\label{exp-neg-eps}`
    exp_neg_eps = RBig.try_from(exp_neg_eps)

    fixed_point = (RBig(1) - delta) / (RBig(1) + exp_eps)

    if fixed_point >= RBig(1, 2):
        raise ValueError("fixed point of tradeoff curve must be less than 1/2")

    def tradeoff(alpha: RBig) -> RBig:  # `\label{tradeoff}`
        t1 = RBig(1) - delta - exp_eps * alpha
        t2 = exp_neg_eps * (RBig(1) - delta - alpha)
        return max(max(t1, t2), RBig(0))

    return tradeoff, fixed_point



# type: ignore
def make_is_equal(
    input_domain: VectorDomain[AtomDomain[TIA]], 
    input_metric: M, 
    value: TIA
): # |\label{line:def}|
    output_row_domain = atom_domain(T=bool)

    def is_equal(arg: TA) -> TA: # |\label{line:function}|
        return value == arg

    return make_row_by_row( # |\label{line:row-by-row}|
        input_domain, 
        input_metric, 
        output_row_domain, 
        is_equal
    )


# type: ignore
def quantile_cnd(
    uniform: RBig, f: Callable[[RBig], RBig], c: RBig
) -> RBig | None:
    if uniform < c:
        return quantile_cnd(RBig(1) - f(uniform), f, c) - RBig(1)
    elif uniform <= RBig(1) - c:  # the linear function
        num = uniform - RBig(1, 2)
        den = RBig(1) - RBig(2) * c
        if den.is_zero():
            return
        return num / den
    else:
        return quantile_cnd(f(RBig(1) - uniform), f, c) + RBig(1)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Literal, Optional, Any, Iterator, cast, get_args, TYPE_CHECKING
from functools import reduce
from math import sqrt
from opendp._internal import (
    _extrinsic_distance,
    _extrinsic_domain,
    _make_transformation,
    _new_pure_function,
)
from opendp.combinators import make_composition
from opendp.domains import atom_domain, vector_domain
from opendp.extras._utilities import to_then
from opendp.measurements import then_noise
from opendp.measures import max_divergence, zero_concentrated_divergence
from opendp.metrics import (
    _get_bound,
    frame_distance,
    l1_distance,
    l2_distance,
    symmetric_distance,
)
from opendp.mod import (
    ApproximateDivergence,
    AtomDomain,
    Domain,
    ExtrinsicDistance,
    ExtrinsicDomain,
    FrameDistance,
    LazyFrameDomain,
    Measure,
    Measurement,
    Metric,
    Transformation,
)

if TYPE_CHECKING:  # pragma: no cover
    from opendp.extras.polars import Bound

@dataclass
class Count:
    """Denotes a count query."""

    by: tuple[str, ...]
    """Columns to group by."""
    weight: float = 1.0
    """Importance of this count query. 
    
    - Used by AIM to prioritize cliques.
    - Used by Fixed to distribute privacy budget.
    """

    def __post_init__(self):
        self.by = tuple(self.by)
        if self.weight < 0:
            raise ValueError(f"weight ({self.weight}) must be non-negative")


def mirror_descent(
    domain,  # mbi.Domain
    loss_fn,  # marginal_loss.MarginalLossFn | list[LinearMeasurement]
    *,
    potentials=None,  # CliqueVector | None
):  # MarkovRandomField
    """Fit a MarkovRandomField over the domain and loss function using mirror descent.

    Replicate the API of this function to `use other optimizers from Private-PGM <https://private-pgm.readthedocs.io/en/latest/_autosummary_output/mbi.estimation.html#module-mbi.estimation>`_.
    """
    from mbi.estimation import mirror_descent  # type: ignore[import-untyped,import-not-found]

    return mirror_descent(domain, loss_fn, potentials=potentials)

OnewayType = Literal["all", "unkeyed"]
ONEWAY_ALL, ONEWAY_UNKEYED = get_args(OnewayType)

@dataclass(kw_only=True, frozen=True)
class Algorithm(ABC):
    """Base class for configuration of contingency table algorithms."""

    estimator: Callable = mirror_descent
    """Optimizer to use to fit a MarkovRandomField.

    Defaults to :py:func:`opendp.extras.mbi.mirror_descent`.
    Any function matching the signature of ``mirror_descent`` 
    can be used to customize how the MarkovRandomField is optimized/estimated.
    See `mbi.estimation <https://private-pgm.readthedocs.io/en/latest/_autosummary_output/mbi.estimation.html>`_ for other optimizers.
    """
    oneway: OnewayType = ONEWAY_ALL
    """Fit one-way marginals for all columns, or only unkeyed columns."""
    oneway_split: Optional[float] = None
    """Proportion of budget to use for oneway release.
    
    When ``oneway_split`` is not set, defaults to half of the budget.

    If oneway is ``unkeyed``, budget is further reduced by the proportion of columns with missing keys or cuts.
    That is, when all columns have keys, then ``oneway_split`` is zero,
    and when no columns have keys, then ``oneway_split`` is one-half.
    """

    def __post_init__(self):
        if self.oneway not in get_args(OnewayType):
            raise ValueError(f'oneway ({self.oneway}) must be in {get_args(OnewayType)}')

        if self.oneway_split is not None and not (0 <= self.oneway_split < 1):
            raise ValueError(f"oneway_split ({self.oneway_split}) must be in [0, 1)")

    @abstractmethod
    def make_marginals(
        self,
        input_domain: LazyFrameDomain,
        input_metric: FrameDistance,
        output_measure: Measure,
        d_in: list["Bound"],
        d_out: float,
        *,
        marginals: dict[tuple[str, ...], Any],
        model,  # MarkovRandomField
    ) -> Measurement: ...


def typed_dict_domain(domains: dict[Any, Domain]) -> ExtrinsicDomain:
    """Domain containing a fixed-key dictionary of elements, each with its own domain."""

    def _member(x):
        if not isinstance(x, dict):
            raise ValueError("data must be a dict")
        if set(domains.keys()) != set(x.keys()):
            raise ValueError("data must share key-set with domain")
        return all(domains[k].member(x[k]) for k in domains)

    ident = ", ".join(f"{k}: {str(d)}" for k, d in domains.items())
    return _extrinsic_domain(
        identifier=f"TypedDictDomain({ident})",
        member=_member,
        descriptor=TypedDictDomain(domains),
    )


class TypedDictDomain(dict[tuple[str, ...], Domain]):
    pass


def typed_dict_distance(inner_metric: Metric) -> ExtrinsicDistance:
    """Dictionary distance metric.

    The metric forms a valid metric space when paired with a typed dict domain
    where the inner metric can be applied on values.
    The metric computes the distances between values in each key-value pair of the dictionary.
    """
    return _extrinsic_distance(
        identifier=f"TypedDictDistance({inner_metric})",
        descriptor=TypedDictDistance(inner_metric),
    )


@dataclass(frozen=True)
class TypedDictDistance:
    inner_metric: Metric


def get_std(measure: Measure, scale: float) -> float:
    if isinstance(measure, ApproximateDivergence):
        measure = measure.inner_measure

    if measure == max_divergence():
        return scale * sqrt(2)
    if measure == zero_concentrated_divergence():
        return scale
    message = f"output_measure ({measure}) must be max_divergence() or zero_concentrated_divergence()"
    raise ValueError(message)


def get_associated_metric(measure: Measure) -> Metric:
    if measure == max_divergence():
        return l1_distance(T="u32")
    if measure == zero_concentrated_divergence():
        return l2_distance(T="u32")
    message = f"output_measure ({measure}) must be max_divergence() or zero_concentrated_divergence()"
    raise ValueError(message)


def prior(x: float) -> float:
    """Returns the next smaller (prior) float"""
    import numpy as np

    return np.nextafter(x, -np.inf)


def get_cardinalities(input_domain: LazyFrameDomain) -> dict[str, int]:
    """Retrieves the cardinalities of each column in a lazyframe domain"""

    if not isinstance(input_domain, LazyFrameDomain):
        raise ValueError("input_domain must be dp.LazyFrameDomain")

    def get_cardinality(col):
        element_domain = input_domain.get_series_domain(col).element_domain
        if not isinstance(element_domain, AtomDomain):
            raise ValueError("input_domain columns must contain atomic data")

        bounds = element_domain.bounds
        if bounds is None:
            raise ValueError("input_domain columns must be bounded")
        lower, upper = bounds
        if lower != 0:
            raise ValueError("input_domain columns must be lower bounded by zero")
        return upper + 1

    return {col: get_cardinality(col) for col in input_domain.columns}


def make_stable_marginals(
    input_domain: LazyFrameDomain,
    input_metric: FrameDistance,
    output_inner_metric: Metric,
    cliques: list[tuple[str, ...]],
) -> Transformation:
    """Return a transformation that computes all marginals in a workload."""
    from opendp.extras.numpy import arrayd_domain
    from opendp.extras.polars import Bound
    import polars as pl  # type: ignore[import-not-found]
    import numpy as np  # type: ignore[import-not-found]

    if input_metric != frame_distance(symmetric_distance()):
        message = f"input_metric ({input_metric}) must be frame_distance(symmetric_distance())"
        raise ValueError(message)

    cardinalities = get_cardinalities(input_domain)

    metrics = {l1_distance(T="u32"): 1, l2_distance(T="u32"): 2}
    if output_inner_metric not in metrics:
        message = f"inner_output_metric ({output_inner_metric}) must be in {set(metrics.keys())}"
        raise ValueError(message)
    p = metrics[output_inner_metric]

    def shape(by: tuple[str, ...]) -> tuple[int, ...]:
        return tuple(cardinalities[c] for c in by)

    def pivot(x, clique: tuple[str, ...]):
        y = np.zeros(shape(clique), dtype=np.int32)
        y[tuple(x[clique].to_numpy().T)] = x["len"].to_numpy()
        return y

    def function(data):
        dfs = pl.collect_all([data.group_by(c).agg(pl.len()) for c in cliques])
        return {c: pivot(m, c) for m, c in zip(dfs, cliques)}

    def count_sensitivity(bounds: list[Bound], clique: tuple[str, ...]) -> float:
        l1 = _get_bound(bounds, []).per_group

        bound: Bound = _get_bound(bounds, [pl.col(c) for c in clique])
        l0 = (bound.num_groups or l1) ** (1 / p)
        li = bound.per_group or l1
        return min(l1, l0 * li)

    def stability_map(d_in: list[Bound]) -> dict[tuple[str, ...], float]:
        return {clique: count_sensitivity(d_in, clique) for clique in cliques}

    return _make_transformation(
        input_domain,
        input_metric,
        output_domain=typed_dict_domain(
            {c: arrayd_domain(shape=shape(c), T="i32") for c in cliques}
        ),
        output_metric=typed_dict_distance(output_inner_metric),
        function=function,
        stability_map=stability_map,
    )


def make_noise_marginals(
    input_domain: ExtrinsicDomain,
    input_metric: ExtrinsicDistance,
    output_measure: Measure,
    cliques: list[tuple[str, ...]],
    scale: float,
    weights: Optional[list[float]] = None,
) -> Measurement:
    """Make a measurement that releases multiple DP marginals"""
    measurements = [
        make_noise_marginal(
            input_domain, input_metric, output_measure, clique, scale / weight
        )
        for clique, weight in zip(cliques, weights or [1] * len(cliques))
    ]
    return make_composition(measurements)


then_noise_marginals = to_then(make_noise_marginals)


def make_noise_marginal(
    input_domain: ExtrinsicDomain,
    input_metric: ExtrinsicDistance,
    output_measure: Measure,
    clique: tuple[str, ...],
    scale: float,
) -> Measurement:
    """Make a measurement that releases a DP marginal"""
    from opendp.extras.numpy import NPArrayDDomain
    import numpy as np  # type: ignore[import-not-found]
    from mbi import LinearMeasurement  # type: ignore[import-untyped,import-not-found]

    clique_domain = input_domain.cast(TypedDictDomain)[clique]
    inner_metric = input_metric.cast(TypedDictDistance).inner_metric

    clique_domain.cast(NPArrayDDomain)

    associated_metric = get_associated_metric(output_measure)

    if inner_metric != associated_metric:
        message = f"input_metric's inner metric ({inner_metric}) doesn't match the output_measure's associated metric ({associated_metric})"
        raise ValueError(message)

    t_marginal = _make_transformation(
        input_domain,
        input_metric,
        output_domain=vector_domain(atom_domain(T="i32")),
        output_metric=inner_metric,
        function=lambda exact_tabs: exact_tabs[clique].astype(np.int32).flatten(),
        stability_map=lambda d_in: d_in[clique],
    )

    def function(x):
        return LinearMeasurement(x, clique, stddev=get_std(output_measure, scale))

    return (
        t_marginal >> then_noise(output_measure, scale) >> _new_pure_function(function)
    )


def row_major_order(keys: Iterator):
    """Joins a list of keysets that align to row-major ordering."""

    def reducer(a, b):
        return a.join(b, how="cross", maintain_order="left_right")

    return reduce(reducer, (keyset.to_frame() for keyset in keys))


def weight_marginals(
    marginals: dict[tuple[str, ...], Any], *new_marginals
) -> dict[tuple[str, ...], Any]:
    from mbi import LinearMeasurement  # type: ignore[import-untyped,import-not-found]

    marginals = marginals.copy()

    for new_marginal in new_marginals:
        if not isinstance(new_marginal, LinearMeasurement):
            raise ValueError("each new marginal must be of type LinearMeasurement")

        clique = new_marginal.clique
        old_marginal = cast(Optional[LinearMeasurement], marginals.get(clique))

        if old_marginal is None:
            marginals[clique] = new_marginal
            continue

        old_var = old_marginal.stddev**2
        old = old_marginal.noisy_measurement

        new_var = new_marginal.stddev**2
        new = new_marginal.noisy_measurement

        weighted_var = 1 / (1 / old_var + 1 / new_var)
        weighted = (old / old_var + new / new_var) * weighted_var

        marginals[clique] = LinearMeasurement(
            weighted, clique, stddev=sqrt(weighted_var)
        )

    return marginals



# type: ignore
# analogous to impl NoisePrivacyMap<L1Distance<RBig>, MaxDivergence> for ZExpFamily<1> in Rust
class ZExpFamily1:
    def noise_privacy_map(
        self, _input_metric: L1Distance[RBig], _output_measure: MaxDivergence
    ) -> PrivacyMap[L1Distance[RBig], MaxDivergence]:
        scale = self.scale
        if scale < RBig.ZERO:  # |\label{line:neg-scale}|
            raise "scale must be non-negative"

        def privacy_map(d_in: RBig):
            if d_in < RBig.ZERO:  # |\label{line:neg-sens}|
                raise "sensitivity must be non-negative"

            if d_in.is_zero():  # |\label{line:zero-sens}|
                return 0.0

            if scale.is_zero():  # |\label{line:zero-scale}|
                return float("inf")

            return f64.inf_cast(d_in / scale)  # |\label{line:map}|

        return PrivacyMap.new_fallible(privacy_map)



# type: ignore
def find_min_covering(
    must_cover: set[T], sets: list[set[T], u32]
) -> list[tuple[set[T], u32]] | None:
    
    covered = list()  # `\label{covered}`

    while must_cover:  # `\label{loop}`

        def score(pair):
            by, weight = pair
            return len(by & must_cover), -len(by), -weight

        best_match = max(sets.items(), key=score)

        if best_match is None or best_match[0].isdisjoint(must_cover):
            return None
        best_set, weight = best_match

        must_cover -= best_set  # `\label{state}`
        covered[best_set] = weight

    return covered



# type: ignore
def gumbel_top_k(
    x: list[TIA], k: usize, scale: FBig, negate: bool,
) -> list[usize]:
    if scale.is_zero():
        if negate:  # |\label{fn-optimize}|
            cmp = lambda a, b: a < b
        else:
            cmp = lambda a, b: a > b
        
        def max_sample(a, b):  # |\label{fn-max-sample-exact}|
            return a if cmp(a[1], b[1]) else b

        return [i for i, _ in top(x, k, max_sample)]   # |\label{fn-top-exact}|

    if all(w[0] == w[1] for w in windows(x, 2)):
        # All values are the same.  # |\label{fn-all-same}|
        x.shuffle()
        return x[:k]

    def try_cast(v):
        try:
            return FBig.try_from(v)
        except Exception:
            return None

    # Cast to FBig.  # |\label{fn-cast}|
    x = ((i, try_cast(x_i)) for i, x_i in enumerate(x))
    # Discard failed casts.  # |\label{fn-filter-nan}|
    x = ((i, x_i) for i, x_i in x if x_i is not None)

    # Normalize sign.  # |\label{fn-normalize-sign}|
    y = ((i, -x_i if negate else x_i) for i, x_i in x)

    # Initialize partial sample.  # |\label{fn-init-sample}|
    def partial_sample(shift):
        rv = Gumbel(shift, scale)  # |\label{fn-rv}|
        return PartialSample.new(rv)  # |\label{fn-partial-sample}|

    y = ((i, partial_sample(y_i)) for i, y_i in y)

    # Reduce to the k pairs with largest samples.  # |\label{fn-max-sample}|
    def max_sample(a, b):
        return a if a[1].greater_than(b[1]) else b

    y_top = top(y, k, max_sample) # |\label{fn-top}|

    # Discard samples, keep indices.  # |\label{fn-return-indices}|
    return [i for i, _ in y_top]



'''
These checks have grown over time, and off-the-shelf tools have been improving at the same time:
Now that our docs are cleaner, and the tools are also better, it might make sense to drop test_ast.py.
Sphinx/RST isn't supported by all tools, but we still have a couple good options.
Either Docsig and Pydoclint could be used as a flake8 plugin.


# Worth another look!

## Docsig
- https://pypi.org/project/docsig/
- `docsig src`
- Started in 2022.

## Pydoclint
- https://github.com/jsh9/pydoclint
- `pydoclint src --style=sphinx`
- Only started in 2023, so it wasn't always an option for us.


# Not for us

## Ruff
- Does not yet support Sphinx: https://github.com/astral-sh/ruff/issues/6606

## Pydoctest
- Doesn't run out of the box: "No module named '_gdbm'"
- Limited configurability
- Small user base

## Pydocstyle
- https://www.pydocstyle.org/en/stable/
- `pydocstyle src`
- This is more about the style of the text: Capitalization, punctuataion, etc.
- Too prescriptive for us.

## flake8-docstrings-complete
- https://pypi.org/project/flake8-docstrings-complete/
- Doesn't support Sphinx

## darglint2
- https://github.com/akaihola/darglint2
- The package itself says that newer tools are better
- but it does have a good survey of the other tools!
'''


from pathlib import Path
import ast
import re
from typing import NamedTuple

import pytest


def ast_ends_with_ellipsis(node):
    '''
    >>> ast_ends_with_ellipsis(ast.parse('def nope(): pass').body[0])
    False
    >>> ast_ends_with_ellipsis(ast.parse('def yeah(): ...').body[0])
    True

    Need body[0] because parse wraps these in a "Module".
    '''
    last_node_value = getattr(node.body[-1], 'value', None)
    last_node_value_value = getattr(last_node_value, 'value', None)
    return last_node_value_value == Ellipsis


def ast_has_return(tree):
    '''
    >>> ast_has_return(ast.parse('def nope(): pass'))
    False
    >>> ast_has_return(ast.parse('def yeah(): return 1'))
    True
    '''
    for node in ast.walk(tree):
        if isinstance(node, ast.Return):
            return True
    return False


def ast_return_type(tree):
    '''
    >>> ast_return_type(ast.parse('def f() -> list[int]: return [1]').body[0])
    'list[int]'

    Need body[0] because parse wraps these in a "Module".
    '''
    type_node = tree.returns
    if type_node is not None:
        return ast.unparse(type_node)


def check_directive_order(docstring):
    '''
    >>> check_directive_order("""
    ...     :param abc: The input
    ...     :paran xyz: Typo!
    ... """)
    'Unknown directives: :paran'
    
    >>> check_directive_order("""
    ...     :return: The output
    ...     :param xyz: The input
    ... """)
    'Directives :return:param are not in canonical order: (param(type)?)*(return)?(rtype)?(raises)*(example)?'
    '''
    directives = re.findall(r'^\s*(\:\w+:?)', docstring, re.MULTILINE)
    unknown_directives = set(directives) - {':param', ':rtype:', ':type', ':raises', ':example:', ':return:'}
    if unknown_directives:
        return f'Unknown directives: {", ".join(unknown_directives)}'
    
    order = ''.join(re.sub(r':$', '', d) for d in directives)
    # TODO: Has 169 failures if we require ":return" and ":rtype" together:
    # canonical_order = r'^(:param(:type)?)*(:return:rtype)?(:raises)*(:example)?$'

    # TODO: Has 139 failures if we require ":return" if ":rtype" is given:
    # (Low priority: from the type and the function name, the return may be clear enough)
    # canonical_order = r'^(:param(:type)?)*(:return(:rtype)?)?(:raises)*(:example)?$'

    # TODO: Has 28 failures if we require ":rtype" if ":return" is given:
    # canonical_order = r'^(:param(:type)?)*((:return)?:rtype)?(:raises)*(:example)?$'
    
    canonical_order = r'^(:param(:type)?)*(:return)?(:rtype)?(:raises)*(:example)?$'
    if not re.search(canonical_order, order):
        short_order = re.sub(r'[:$^]', '', canonical_order)
        return f'Directives {order} are not in canonical order: {short_order}'


def check_directive_continuity(docstring):
    '''
    TODO: Add support for multi-line, indented param descriptions.

    >>> check_directive_continuity("""
    ...     :param xyz: The input
    ...     Surprise!
    ...     :return: The output
    ... """)
    'Found another directive after non-directive: :return: The output'
    '''
    directives_started = False
    directives_ended = False
    for line in docstring.split('\n'):
        line = line.strip()
        if not line:
            continue
        if line.startswith(':'):
            if directives_ended:
                return f'Found another directive after non-directive: {line}'
            directives_started = True
        elif directives_started:
            directives_ended = True


def check_list_space(docstring):
    '''
    >>> check_list_space("""
    ...     Add a blank line after this one!
    ...     1. One thing
    ...     2. After another
    ... """)
    'Add a blank line above list that begins with: 1. One thing'

    >>> check_list_space("""
    ...     >>> 5.0 - 4.0
    ...     1.0
    ...
    ...     That should pass
    ...     1. but this should not!
    ... """)
    'Add a blank line above list that begins with: 1. but this should not!'

    >>> check_list_space("""
    ...     * This
    ...     * is fine,
    ...
    ...     but this is
    ...     * not ok!
    ... """)
    'Add a blank line above list that begins with: * not ok!'
    '''
    prev_is_text = False
    for line in docstring.split('\n'):
        line = line.strip()
        if prev_is_text and (line.startswith('1.') or line.startswith('* ')):
            return f'Add a blank line above list that begins with: {line}'
        prev_is_text = line and not (
            line.startswith('>>>')
            or line.startswith('...')
            or line.startswith('* ')
        )


class Checker():
    def __init__(self, name, tree, docstring, is_public, is_verbose):
        self.name = name
        self.tree = tree
        self.docstring = docstring
        self.is_public = is_public
        self.is_verbose = is_verbose

        self.errors = []

        args = self.tree.args
        self.all_ast_args = (
            args.posonlyargs
            + args.args
            + args.kwonlyargs
        )
        if args.vararg is not None:
            self.all_ast_args.append(args.vararg)
        if args.kwarg is not None:
            self.all_ast_args.append(args.kwarg)
        if len(self.all_ast_args) and self.all_ast_args[0].arg in ['self', 'cls']:
            # TODO: Confirm that this is a method in a class.
            self.all_ast_args.pop(0)
        
        if self.name == '__init__' and self.docstring:
            self.errors.append('Move docstring up to class')

    def _check_docstring(self):
        checks = [
            check_directive_order,
            check_directive_continuity,
            check_list_space,
        ]
        for check in checks:
            error = check(self.docstring)
            if error:
                self.errors.append(error)

    def _check_params(self):
        doc_param_dict = dict(re.findall(r':param (\w+): *(.*)', self.docstring))
        # TODO: Check even if private function.
        if self.is_public:
            k_missing_v = [k for k, v in doc_param_dict.items() if not v]
            if k_missing_v:
                self.errors.append(f'params missing descriptions: {", ".join(k_missing_v)}')

        ast_arg_names = {arg.arg for arg in self.all_ast_args}
        if doc_param_dict or is_public:
            # Private functions don't need to document params,
            # but if they do, they should be consistent with signature.
            if doc_param_dict.keys() != ast_arg_names:
                self.errors.append(
                    f'docstring params ({", ".join(doc_param_dict.keys())}) '
                    f'!= function signature ({", ".join(ast_arg_names)})'
                )

    def _check_types(self):
        doc_type_dict = (
            # Get all documented parameters, even if there aren't types...
            {name: None for name in re.findall(r':param (\w+):', self.docstring)}
            | {
                # ... and then add types, where documented.
                # Allow either ":py:ref:" or ":ref:".
                # No reason to impose a standard if both work with the tools.
                k: re.sub(r'(?::py)?:ref:`(\w+)`', r'\1', v)
                for k, v in re.findall(r':type (\w+): *(.*)', self.docstring)
            }
        )

        ast_arg_annotation_dict = {
            arg.arg: getattr(arg, 'annotation', None)
            for arg in self.all_ast_args
        }
        ast_type_dict = {
            # "v and ..." ensures that None is passed through.
            # TODO: Has 32 failures if we require parameters with "Optional" types
            # to also have "Optional" in the  docstring.
            # k: v and ast.unparse(v)
            k: v and re.sub(r'Optional\[(.+)\]', r'\1', ast.unparse(v))
            for k, v in ast_arg_annotation_dict.items()
        }
        
        for k in doc_type_dict.keys() | ast_type_dict.keys():
            doc_type = doc_type_dict.get(k)
            ast_type = ast_type_dict.get(k)

            # TODO: Ignoring private functions, there are still 56 functions where 
            # either the documentation type or the annotation type is missing.
            if doc_type is None or ast_type is None:
                continue

            if doc_type != ast_type:
                self.errors.append(
                    f'docstring type ({doc_type}) '
                    f'!= function signature types ({ast_type}) '
                    f'for {k}'
                )

    def _check_return(self):
        has_return_statement = ast_has_return(self.tree)
        has_return_doc = ':return' in self.docstring or ':rtype' in self.docstring
        # TODO: Has 68 failures if "return" statements require either ":return" or ":rtype".
        # if self.is_public and has_return_statement and not has_return_doc:
        #     self.errors.append('return statement, but no :return or :rtype in docstring')
        if has_return_doc and not has_return_statement:
            self.errors.append(':return: directive, but no return statement')

        rtype_match = re.search(r':rtype:(.*)', self.docstring)
        if rtype_match:
            doc_rtype = rtype_match.group(1).strip().replace('"', "'")
            ast_rtype = ast_return_type(self.tree)
            # We trust mypy to check that the annotation is consistent with the actual return,
            # so we just check that the annotation is consistent with docstring.
            if ast_rtype is None:
                self.errors.append(f'to match :rtype:, add "-> {doc_rtype}"')
            # If it's a single character, probably a typevar, which we won't try to resolve.
            elif len(ast_rtype) > 1:
                if doc_rtype != ast_rtype:
                    self.errors.append(f'update :rtype: from "{doc_rtype}" to "{ast_rtype}"')

    def get_errors(self):
        self._check_docstring()
        self._check_params()
        self._check_types()
        self._check_return()
        if not self.errors:
            return
        if self.is_verbose:
            # split across multiple lines for readability:
            return '\n' + '\n'.join(self.errors)
        # else all on one line, though it may be truncated:
        return '; '.join(
            self.errors
            if len(self.errors) == 1
            else [f'({i+1}) {e}' for i, e in enumerate(self.errors)]
        )


PUBLIC = 'public'
PRIVATE = 'private'

class CodeObj(NamedTuple):
    file: str
    name: str
    tree: ast.AST
    visibility: str  # str rather than bool so the pytest report is more readable.


classes = []
functions = []

src_dir_path = Path(__file__).parent.parent / 'src'
for code_path in src_dir_path.glob('**/*.py'):
    is_public = True
    if any(parent.name.startswith('_') for parent in code_path.parents):
        is_public = False
    if code_path.name.startswith('_') and code_path.name != '__init__.py':
        is_public = False
    code = code_path.read_text()
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if not isinstance(node, (ast.FunctionDef, ast.ClassDef)):
            continue
        if ast_ends_with_ellipsis(node):
            continue
        if node.name.startswith('_'):
            is_public = False

        code_obj = CodeObj(
            file=f'{code_path.parent.name}/{code_path.name}',
            name=node.name,
            tree=node,
            visibility=PUBLIC if is_public else PRIVATE,
        )
        if isinstance(node, ast.FunctionDef):
            functions.append(code_obj)
        if isinstance(node, ast.ClassDef):
            classes.append(code_obj)

# Typo check:
assert len(functions) > 100


@pytest.mark.parametrize("file,name,tree,visibility", functions)
def test_function_docs(file, name, tree, visibility, pytestconfig):
    where = f'In {file}, line {tree.lineno}, def {name}'
    is_public = visibility == PUBLIC

    docstring = ast.get_docstring(tree)
    if not is_public and docstring is None:
        return
    assert docstring is not None, f'{where}: add docstring or make private'

    errors = Checker(
        name=name,
        tree=tree,
        docstring=docstring,
        is_public=is_public,
        is_verbose=pytestconfig.getoption("verbose") > 0
    ).get_errors()
    if errors:
        pytest.fail(f'{where}: {errors}')

@pytest.mark.parametrize("file,name,tree,visibility", classes)
def test_class_docs(file, name, tree, visibility):
    where = f'In {file}, line {tree.lineno}, class {name}'
    is_public = visibility == PUBLIC

    docstring = ast.get_docstring(tree)
    if not is_public and docstring is None:
        return
    assert docstring is not None, f'{where}: add docstring or make private'


# type: ignore
# analogous to impl NoiseThresholdPrivacyMap<L02I<AbsoluteDistance<RBig>>, Approximate<ZeroConcentratedDivergence>> for ZExpFamily<2> in Rust
class ZExpFamily2:
    def noise_threshold_privacy_map(
        self,
        _input_metric: L02InfDistance[AbsoluteDistance[RBig]],
        output_measure: Approximate[ZeroConcentratedDivergence],
        threshold: UBig,
    ) -> PrivacyMap[L02InfDistance[AbsoluteDistance[RBig]], Approximate[ZeroConcentratedDivergence]]:
        # |\label{line:noise-privacy-map}|
        noise_privacy_map = self.noise_privacy_map(L2Distance.default(), output_measure[0])
        scale = self.scale

        def privacy_map(d_in: tuple[u32, RBig, RBig]):
            l0, l2, li = d_in

            li_sign, li = li.floor().into_parts() # |\label{line:li-floor}|
            if li_sign != Sign.Positive: # |\label{line:li-check}|
                raise f"l-infinity sensitivity ({li}) must be non-negative"
            
            l2 = l2.min(li * RBig.try_from(f64.from_(l0).inf_sqrt()))  # |\label{line:l2}|
            li = li.min(l2.floor())  # |\label{line:li}|

            if l2.is_zero():  # |\label{line:zero-sens}|
                return 0.0, 0.0

            if scale.is_zero(): # |\label{line:zero-scale}|
                return f64.INFINITY, 1.0

            rho = noise_privacy_map.eval(l2) # |\label{line:rho}|
            
            if li > threshold: # |\label{line:threshold-check}|
                raise f"threshold must not be smaller than {li}"
            
            d_instability = threshold - li # |\label{line:distance-to-instability}|

            delta_single = conservative_discrete_gaussian_tail_to_alpha(
                scale,
                d_instability,
            )

            delta_joint: f64 = (1.0).inf_sub(
                (1.0).neg_inf_sub(delta_single).neg_inf_powi(IBig.from_(l0)),
            )

            # delta is only sensibly at most 1
            return rho, delta_joint.min(1.0)

        return PrivacyMap.new_fallible(privacy_map)



'''
This module requires extra installs: ``pip install 'opendp[scikit-learn]'``

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.sklearn.decomposition``.    

See also our :ref:`tutorial on diffentially private PCA <dp-pca>`.
'''

from __future__ import annotations
from typing import NamedTuple, Optional, TYPE_CHECKING, Sequence
from opendp.extras.numpy import then_np_clamp
from opendp.context import register
from opendp.extras._utilities import to_then
from opendp.extras.numpy._make_np_mean import make_private_np_mean
from opendp.extras.sklearn._make_eigendecomposition import then_private_np_eigendecomposition
from opendp.mod import Domain, Measurement, Metric
from opendp._lib import import_optional_dependency
from opendp._internal import _make_measurement, _make_transformation

if TYPE_CHECKING: # pragma: no cover
    import numpy # type: ignore[import-not-found]


class PCAEpsilons(NamedTuple):
    '''
    Tuple used to describe the ε-expenditure per changed record in the input data
    '''
    eigvals: float
    eigvecs: Sequence[float]
    mean: Optional[float]


PCAEpsilons.eigvals.__doc__ = 'ε-expenditure to estimate the eigenvalues'
PCAEpsilons.eigvecs.__doc__ = 'ε-expenditure to estimate the eigenvectors'
PCAEpsilons.mean.__doc__ = ''  """ε-expenditure to estimate the mean.

A portion of the budget is used to estimate the mean because the OpenDP PCA algorithm 
releases an eigendecomposition of the sum of squares and cross-products matrix (SSCP), 
not of the covariance matrix. 
If the data is centered beforehand (either by a prior from the user or by privately estimating the mean and then centering), 
then PCA will correspond to the covariance matrix, as expected, 
because the SSCP matrix of centered data is equivalent to a scaled covariance matrix.

If the data is not centered (or the mean is poorly estimated), then the first eigenvector will be dominated by the true mean.
"""


def make_private_pca(
    input_domain: Domain,
    input_metric: Metric,
    unit_epsilon: float | PCAEpsilons,
    norm: float | None = None,
    num_components=None,
) -> Measurement:
    """Construct a Measurement that returns the data mean, singular values and right singular vectors.

    :param input_domain: instance of `array2_domain(size=_, num_columns=_)`
    :param input_metric: instance of `symmetric_distance()`
    :param unit_epsilon: ε-expenditure per changed record in the input data
    :param norm: clamp each row to this norm bound
    :param num_components: optional, number of eigenvectors to release. defaults to num_columns from input_domain

    :return: a Measurement that computes a tuple of (mean, S, Vt)
    """
    import opendp.prelude as dp
    np = import_optional_dependency('numpy')

    dp.assert_features("contrib", "floating-point")

    class PCAResult(NamedTuple):
        mean: numpy.ndarray
        S: numpy.ndarray
        Vt: numpy.ndarray

    input_desc = input_domain.descriptor
    if input_desc.size is None:
        raise ValueError("input_domain's size must be known")  # pragma: no cover

    if input_desc.num_columns is None:
        raise ValueError("input_domain's num_columns must be known")  # pragma: no cover

    if input_desc.p not in {None, 2}:
        raise ValueError("input_domain's norm must be an L2 norm")  # pragma: no cover

    if input_desc.num_columns < 1:
        raise ValueError("input_domain's num_columns must be >= 1")  # pragma: no cover

    num_components = (
        input_desc.num_columns if num_components is None else num_components
    )

    if isinstance(unit_epsilon, float):
        num_eigvec_releases = min(num_components, input_desc.num_columns - 1)
        unit_epsilon = _split_pca_epsilon_evenly(
            unit_epsilon, num_eigvec_releases, estimate_mean=input_desc.origin is None
        )

    if not isinstance(unit_epsilon, PCAEpsilons):
        raise ValueError("epsilon must be a float or instance of PCAEpsilons")  # pragma: no cover

    eigvals_epsilon, eigvecs_epsilons, mean_epsilon = unit_epsilon

    def _eig_to_SVt(decomp):
        eigvals, eigvecs = decomp
        return np.sqrt(np.maximum(eigvals, 0))[::-1], eigvecs.T

    def _make_eigdecomp(norm, origin):
        return (
            (input_domain, input_metric)
            >> then_np_clamp(norm, p=2, origin=origin)
            >> then_center()
            >> then_private_np_eigendecomposition(eigvals_epsilon, eigvecs_epsilons)
            >> (lambda out: PCAResult(origin, *_eig_to_SVt(out)))
        )

    if input_desc.norm is not None:
        if mean_epsilon is not None:
            raise ValueError("mean_epsilon should be zero because origin is known")  # pragma: no cover
        norm = input_desc.norm if norm is None else norm
        norm = min(input_desc.norm, norm)
        return _make_eigdecomp(norm, input_desc.origin)
    elif norm is None:
        raise ValueError("must have either bounded `input_domain` or specify `norm`")  # pragma: no cover


    # make releases under the assumption that d_in is 2.
    unit_d_in = 2

    compositor = dp.c.make_adaptive_composition(
        input_domain,
        input_metric,
        dp.max_divergence(),
        d_in=unit_d_in,
        d_mids=[mean_epsilon, _make_eigdecomp(norm, 0).map(unit_d_in)],
    )

    def _function(data):
        nonlocal input_domain
        qbl = compositor(data)

        # find origin
        m_mean = dp.binary_search_chain(  # type: ignore[misc]
            lambda s: make_private_np_mean(
                input_domain, input_metric, s, norm=norm, p=1
            ),
            d_in=unit_d_in,
            d_out=mean_epsilon,
            T=float,
        )
        origin = qbl(m_mean)
        # make full release
        return qbl(_make_eigdecomp(norm, origin))

    return _make_measurement(
        input_domain,
        input_metric,
        compositor.output_measure,
        _function,
        compositor.map,
    )


# generate then variant of the constructor
then_private_pca = to_then(make_private_pca)
register(make_private_pca)


class PCA:
    '''
    DP wrapper for `sklearn's PCA <https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html>`_.
    This implementation is based on `Differentially Private Covariance Estimation <https://papers.nips.cc/paper_files/paper/2019/hash/4158f6d19559955bae372bb00f6204e4-Abstract.html>`_ by Kareem Amin, et al.

    Trying to create an instance without sklearn installed will raise an ``ImportError``.
    
    See the :ref:`tutorial on diffentially private PCA <dp-pca>` for details.

    :param whiten: Mirrors the corresponding sklearn parameter:
        When ``True`` (``False`` by default) the ``components_`` vectors are multiplied
        by the square root of n_samples and then divided by the singular values
        to ensure uncorrelated outputs with unit component-wise variances.
    '''
    def __init__(
        self,
        *,
        epsilon: float,
        row_norm: float,
        n_samples: int,
        n_features: int,
        n_components: int | float | str | None = None,
        n_changes: int = 1,
        whiten: bool = False,
    ) -> None:  # pragma: no cover
        # Error if constructor called without dependency:
        import_optional_dependency('sklearn.decomposition')
        # used for mypy typing
        self.n_samples_ = None
        self.components_ = None
        self.n_components_ = None
        self.explained_variance_ = None
        self.explained_variance_ratio_ = None
        self.singular_values_ = None

    @property
    def n_features(self):
        '''
        Number of features
        '''
        ...

    def fit(self, X, y=None):
        '''
        Fit the model with X.

        :param X: Training data, where ``n_samples`` is the number of samples and ``n_features`` is the number of features.
        :param y: Ignored
        '''
        ...

    # this overrides the scikit-learn method to instead use the opendp-core constructor
    def _fit(self, X):
        ...

    def _prepare_fitter(self) -> Measurement:  # type: ignore[empty-body]
        """Returns a measurement that computes the mean and eigendecomposition,
        and then apply those releases to self."""
        ...

    def _postprocess(self, values):
        """A function that applies a release of the mean and eigendecomposition to self"""
        ...

    def measurement(self) -> Measurement:  # type: ignore[empty-body]
        """Return a measurement that releases a fitted model."""
        ...

    def _validate_params(*args, **kwargs):
        ...


_decomposition = import_optional_dependency('sklearn.decomposition', False)
if _decomposition is not None:
    class PCA(_decomposition.PCA):  # type: ignore  # noqa: F811
        def __init__(
            self,
            *,
            epsilon: float,
            row_norm: float,
            n_samples: int,
            n_features: int,
            n_components: int | float | str | None = None,
            n_changes: int = 1,
            whiten: bool = False,
        ) -> None:
            super().__init__(
                n_components or n_features,
                whiten=whiten,
            )
            self.epsilon = epsilon
            self.row_norm = row_norm
            self.n_samples = n_samples
            self.n_features_in_ = n_features
            self.n_changes = n_changes

        @property
        def n_features(self):
            '''
            Number of features
            '''
            return self.n_features_in_

        # This isn't strictly necessary, since we just call the superclass method,
        # but this lets us document a frequently used method,
        # and avoids a number of mypy warnings.
        def fit(self, X, y=None):
            '''
            Fit the model with X.

            :param X: Training data, where ``n_samples`` is the number of samples and ``n_features`` is the number of features.
            :param y: Ignored
            '''
            return super().fit(X)

        # this overrides the scikit-learn method to instead use the opendp-core constructor
        def _fit(self, X):
            return self._prepare_fitter()(X)

        def _prepare_fitter(self) -> Measurement:
            """Returns a measurement that computes the mean and eigendecomposition,
            and then apply those releases to self."""
            import opendp.prelude as dp

            if hasattr(self, "components_"):
                raise ValueError("DP-PCA model has already been fitted")  # pragma: no cover

            input_domain = dp.numpy.array2_domain(
                num_columns=self.n_features_in_, size=self.n_samples, T=float
            )
            input_metric = dp.symmetric_distance()

            n_estimated_components = (
                self.n_components
                if isinstance(self.n_components, int)
                else self.n_features_in_
            )

            return make_private_pca(
                input_domain,
                input_metric,
                self.epsilon / self.n_changes * 2,
                norm=self.row_norm,
                num_components=n_estimated_components,
            ) >> self._postprocess

        def _postprocess(self, values):
            """A function that applies a release of the mean and eigendecomposition to self"""
            np = import_optional_dependency('numpy')
            from sklearn.utils.extmath import stable_cumsum, svd_flip # type: ignore[import]
            from sklearn.decomposition._pca import _infer_dimension # type: ignore[import]

            self.mean_, S, Vt = values
            U = Vt.T
            n_samples, n_features = self.n_samples, self.n_features_in_
            n_components = self.n_components

            # CODE BELOW THIS POINT IS FROM SKLEARN
            # flip eigenvectors' sign to enforce deterministic output
            U, Vt = svd_flip(U, Vt)

            components_ = Vt

            # Get variance explained by singular values
            explained_variance_ = (S**2) / (n_samples - 1)
            total_var = np.sum(explained_variance_)
            explained_variance_ratio_ = explained_variance_ / total_var
            singular_values_ = S # Store the singular values. 

            # Postprocess the number of components required
            if n_components == "mle":
                n_components = _infer_dimension(explained_variance_, n_samples)
            elif 0 < n_components < 1.0:
                # number of components for which the cumulated explained
                # variance percentage is superior to the desired threshold
                # side='right' ensures that number of features selected
                # their variance is always greater than n_components float
                # passed. More discussion in issue: https://github.com/scikit-learn/scikit-learn/pull/15669
                explained_variance_ratio_np = explained_variance_ratio_
                ratio_cumsum = stable_cumsum(explained_variance_ratio_np)
                n_components = np.searchsorted(ratio_cumsum, n_components, side="right") + 1

            # Compute noise covariance using Probabilistic PCA model
            # The sigma2 maximum likelihood (cf. eq. 12.46)
            if n_components < min(n_features, n_samples):
                self.noise_variance_ = np.mean(explained_variance_[n_components:])
            else:
                self.noise_variance_ = 0.0

            self.n_samples_ = n_samples
            self.components_ = components_[:n_components, :]
            self.n_components_ = n_components
            self.explained_variance_ = explained_variance_[:n_components]
            self.explained_variance_ratio_ = explained_variance_ratio_[:n_components]
            self.singular_values_ = singular_values_[:n_components]

            return U, S, Vt

        def measurement(self) -> Measurement:
            """Return a measurement that releases a fitted model."""
            return self._prepare_fitter() >> (lambda _: self)

        # overrides an sklearn method
        def _validate_params(*args, **kwargs):
            pass



        
def _smaller(v):
    """returns the next non-negative float closer to zero"""
    np = import_optional_dependency('numpy')

    if v < 0:
        raise ValueError("expected non-negative value")  # pragma: no cover
    return v if v == 0 else np.nextafter(v, -1)


def _split_pca_epsilon_evenly(unit_epsilon, num_eigvec_releases, estimate_mean=False):
    num_queries = 3 if estimate_mean else 2
    per_query_epsilon = unit_epsilon / num_queries
    per_evec_epsilon = per_query_epsilon / num_eigvec_releases

    # use conservatively smaller budgets to prevent totals from exceeding total epsilon
    return PCAEpsilons(
        eigvals=per_query_epsilon,
        eigvecs=[_smaller(per_evec_epsilon)] * num_eigvec_releases,
        mean=_smaller(per_query_epsilon) if estimate_mean else None,
    )


def _make_center(input_domain, input_metric):
    import opendp.prelude as dp
    np = import_optional_dependency('numpy')

    dp.assert_features("contrib", "floating-point")

    input_desc = input_domain.descriptor

    kwargs = input_desc._asdict() | {"origin": np.zeros(input_desc.num_columns)}
    return _make_transformation(
        input_domain,
        input_metric,
        dp.numpy.array2_domain(**kwargs),
        input_metric,
        lambda arg: arg - input_desc.origin,
        lambda d_in: d_in,
    )


then_center = to_then(_make_center)


"""
This module requires extra installs: ``pip install 'opendp[polars]'``

For convenience, all the members of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp

The members of this module will then be accessible at ``dp.polars.contingency_table``.
"""

from typing import Optional, Type, Union
from opendp.context import Query
from opendp.extras.mbi._table import ContingencyTable


class ContingencyTableQuery(Query):
    """Extends the Query class to include special attributes for contingency tables.

    :param scale: noise scale to be added to one-way marginals
    :param threshold: cutoff for small groups to be merged into a null group
    :param kwargs: arguments for parent class
    """

    def __init__(
        self,
        *,
        oneway_scale: Optional[float],
        oneway_threshold: Optional[int],
        **kwargs
    ):
        super().__init__(**kwargs)
        self.oneway_scale = oneway_scale
        self.oneway_threshold = oneway_threshold

    def release(
        self,
        data=None,
        bounds: Optional[tuple[float, float]] = None,
        T: Optional[Union[Type[float], Type[int]]] = None,
    ) -> ContingencyTable:
        """Release the ContingencyTable."""
        return super().release(data, bounds, T)



# type: ignore
def then_deintegerize_hashmap(k: i32) -> Function[HashMap[TK, IBig], HashMap[TK, TV]]:
    if k == i32.MIN: # |\label{line:check-k}|
        raise ValueError("k must not be i32.MIN")

    def value_function(v_i):
        return TV.from_rational(x_mul_2k(RBig.from_(v_i), k))

    return Function.new(lambda x: {k_i: value_function(v_i) for k_i, v_i in x.items()})



# type: ignore
class InverseCDF(CanonicalRV):
    # type Edge = RBig

    def inverse_cdf(self, uniform: RBig, _refinements: usize, _R) -> RBig | None:
        f_inv = quantile_cnd(uniform, self.tradeoff, self.fixed_point)  # `\label{f_inv}`
        return f_inv * self.scale + self.shift



# type: ignore
def make_stable_truncate(
    input_domain: DslPlanDomain,
    input_metric: FrameDistance[SymmetricIdDistance],
    plan: DslPlan,
) -> Transformation[
    DslPlanDomain,
    DslPlanDomain,
    FrameDistance[SymmetricIdDistance],
    FrameDistance[SymmetricDistance],
]:
    # the identifier is protected from changes, so we can use the identifier from the input metric
    # instead of the identifier from the middle_metric to match truncations
    input, truncations, truncation_bounds = match_truncations(
        plan, input_metric[0].identifier
    )

    if truncations.is_empty():
        return ValueError("failed to match truncation")

    t_prior = input.make_stable(input_domain, input_metric)
    middle_domain, middle_metric = t_prior.output_space()

    for bound in truncation_bounds:
        for key in bound.by:
            # raises if the key is not infallible row-by-row
            make_stable_expr(  # |\label{line:stable-keys}|
                WildExprDomain(
                    columns=middle_domain.series_domains,
                    context=Context.RowByRow,
                ),
                PartitionDistance(middle_metric[0]),
                key,
            )

    output_domain = middle_domain.clone()
    for truncation in truncations: # |\label{line:truncation}|
        output_domain = truncate_domain(output_domain, truncation)

    def function(plan: DslPlan) -> DslPlan:
        for truncation in truncations:
            match truncation:
                case Truncation.Filter(predicate):
                    plan = DslPlan.Filter(
                        input=plan,
                        predicate=predicate,
                    )

                case Truncation.GroupBy(keys, aggs):
                    plan = DslPlan.GroupBy(
                        input=plan,
                        keys=keys,
                        aggs=aggs,
                        apply=None,
                        maintain_order=False,
                        options=GroupbyOptions.default(),
                    )
        return plan

    def stability_map(id_bounds: Bounds) -> Bounds:
        # |\label{line:total-ids}|
        total_num_ids = id_bounds.get_bound({}).per_group

        # each truncation is used to derive row bounds
        new_bounds = []
        for truncation_bound in truncation_bounds: # |\label{line:truncation-bound}|
            # each truncation is used to derive row bounds
            new_bounds.append(
                truncate_id_bound( # |\label{line:truncate-id-bound}|
                    id_bounds.get_bound(truncation_bound.by), # |\label{line:id-bound-by}|
                    truncation_bound,
                    total_num_ids,
                )
            )
        return Bounds(new_bounds)

    t_truncate = Transformation.new(
        middle_domain,
        output_domain,
        Function.new(function),
        middle_metric,
        FrameDistance(SymmetricDistance),
        StabilityMap.new_fallible(stability_map),
    )
    return t_prior >> t_truncate



# type: ignore
class CompositionMeasure(RenyiDivergence):
    def composability(  # |\label{line:composability}|
        self, adaptivity: Adaptivity
    ) -> Composability:
        return Composability.Concurrent

    def compose(self, d_mids: Vec[Self_Distance]) -> Self_Distance:
        def curve(alpha: float) -> float: # |\label{line:curve}|
            epsilons = [d_mid(alpha) for d_mid in d_mids]

            d_out = 0.0
            for d_mid in epsilons:
                d_out = d_out.inf_add(d_mid)
            return d_out

        return Function.new_fallible(curve)
        



from typing import Sequence, Union, cast, MutableMapping
from inspect import signature

from opendp._lib import *
from opendp.mod import (
    ApproximateDivergence,
    ChangeOneIdDistance,
    Domain,
    ExtrinsicDistance,
    ExtrinsicDomain,
    LazyFrameDomain,
    Measure,
    Metric,
    FrameDistance,
    SeriesDomain,
    SymmetricIdDistance,
    UnknownTypeException,
    Transformation,
    Measurement,
    PrivacyProfile,
    Queryable,
    OdometerQueryable,
    Function,
    AtomDomain,
    OptionDomain,
    VectorDomain,
)
from opendp.typing import RuntimeType, RuntimeTypeDescriptor

ATOM_MAP = {
    'f32': ctypes.c_float,
    'f64': ctypes.c_double,
    'u8': ctypes.c_uint8,
    'u16': ctypes.c_uint16,
    'u32': ctypes.c_uint32,
    'u64': ctypes.c_uint64,
    'i8': ctypes.c_int8,
    'i16': ctypes.c_int16,
    'i32': ctypes.c_int32,
    'i64': ctypes.c_int64,
    'usize': ctypes.c_size_t,
    'bool': ctypes.c_bool,
    'AnyMeasurementPtr': Measurement,
    'AnyTransformationPtr': Transformation,
}

def c_int_limits(type_name):
    c_int_type = ATOM_MAP[type_name]
    signed = c_int_type(-1).value < c_int_type(0).value
    bit_size = ctypes.sizeof(c_int_type) * 8
    signed_limit = 2 ** (bit_size - 1)
    return (-signed_limit, signed_limit - 1) if signed else (0, 2 * signed_limit - 1)

INT_SIZES = {
    ty: c_int_limits(ty) for ty in (
        'u8', 'u16', 'u32', 'u64', 'i8', 'i16', 'i32', 'i64', 'usize',
    )
}
_ERROR_URL_298 = "https://github.com/opendp/opendp/discussions/298"


def _check_and_cast_scalar(expected, value):
    '''
    1. Converts integer value to float if expected value is float
    2. Checks that value is roughly a member of the same data type as expected
    3. Checks that integers are representable at the given data type
    '''
    # relax checks in the case of an int
    if isinstance(value, int) and expected in ["f32", "f64"] and not isinstance(value, bool):
        return float(value)

    inferred = str(RuntimeType.infer(value))

    if expected not in ATOM_EQUIVALENCE_CLASSES.get(inferred, [inferred]):
        raise TypeError(f"inferred type is {inferred}, expected {expected}. See {_ERROR_URL_298}")
    
    if expected in INT_SIZES:
        check_c_int_cast(value, expected)

    return value


def check_c_int_cast(v, type_name):
    lower, upper = INT_SIZES[type_name]
    if not (lower <= v <= upper):
        raise ValueError(f"{v} is not representable by {type_name}")


def py_to_c(value: Any, c_type, type_name: RuntimeTypeDescriptor = None) -> Any:
    """Map from Python `value` to ctypes `c_type`.

    :param value: value to convert to c_type
    :param c_type: expected ctypes type to convert to. Untyped because there is no public superclass.
    :param type_name: optional. rust_type to check inferred rust_type of value against, and/or specify bit depth
    :return: value converted to ctypes representation
    """
    if isinstance(type_name, str):
        type_name = RuntimeType.parse(type_name)

    if isinstance(value, c_type):
        return value

    if c_type == CallbackFnPtr:
        return _wrap_py_func(value, type_name)

    if c_type == TransitionFnPtr:
        return _wrap_py_transition(value, type_name)

    if c_type == ExtrinsicObjectPtr:
        # since the memory is allocated by python, 
        #    don't actually return an ExtrinsicObjectPtr, 
        #    which would call rust to free the Python-allocated ExtrinsicObject
        return ctypes.pointer(ExtrinsicObject(ctypes.py_object(value)))

    # check that the type name is consistent with the value
    if type_name is not None:
        # exit early with a null pointer if trying to load an Option type with a None value
        if isinstance(type_name, RuntimeType) and type_name.origin == "Option":
            if value is None:
                return
            type_name = type_name.args[0]

    if c_type == ctypes.c_void_p:
        if type_name is None:
            raise ValueError("type_name must be known")  # pragma: no cover

        rust_type = str(type_name)

        value = _check_and_cast_scalar(rust_type, value)

        if rust_type in ATOM_MAP:
            return ctypes.byref(ATOM_MAP[rust_type](value))

        if rust_type == "String":
            return ctypes.c_char_p(value.encode())

        raise UnknownTypeException(rust_type)  # pragma: no cover

    if c_type == AnyObjectPtr:
        if isinstance(value, ctypes.POINTER(AnyObject)):
            return value
        
        from opendp._data import slice_as_object
        return slice_as_object(value, type_name) # type: ignore[arg-type]

    if c_type == FfiSlicePtr:
        if type_name is None:
            raise ValueError("type_name must be known")  # pragma: no cover
        return _py_to_slice(value, RuntimeType.parse(type_name))

    if isinstance(value, RuntimeType):
        value = str(value)

    if isinstance(value, str):
        value = value.encode()

    if not isinstance(value, c_type):
        # throw an error if the value is already a c_type, but the wrong one
        # (like passing a Metric into an argument expecting a Domain)
        if hasattr(value, "_type_"):
            raise ValueError(f"Cannot convert {value} to {c_type}")  # pragma: no cover
        value = c_type(value)

    return value


def c_to_py(value: Any) -> Any:
    """Map from ctypes `value` to Python value.
    It is assumed that the C type is simpler than in py_to_c, as the library returns fewer types.

    :param value: data in ctypes format
    :return: copy of data in Python representation
    """
    if isinstance(value, ctypes.POINTER(AnyObject)):
        from opendp._data import object_type, object_as_slice, slice_free

        obj_type = object_type(value)

        if obj_type == PrivacyProfile.__name__:
            return PrivacyProfile(value)
        
        if obj_type == "AnyOdometerQueryable":
            return OdometerQueryable(value)

        if obj_type == "AnyQueryable":
            from opendp.core import queryable_query_type

            query_type = RuntimeType.parse(queryable_query_type(value))

            if query_type == "OnceFrameQuery":
                from opendp.extras.polars import OnceFrame

                return OnceFrame(value)

            return Queryable(value, query_type)

        ffi_slice = object_as_slice(value)
        try:
            return _slice_to_py(ffi_slice, RuntimeType.parse(obj_type))
        finally:
            slice_free(ffi_slice)

    if isinstance(value, ctypes.c_char_p):
        from opendp._data import str_free
        assert value.value is not None
        value_contents = value.value.decode()
        str_free(value)
        return value_contents

    if isinstance(value, BoolPtr):
        from opendp._data import bool_free
        value_contents = value.contents.value
        bool_free(value)
        return value_contents

    if isinstance(value, ctypes.POINTER(ExtrinsicObject)):
        return value.contents.ptr

    if isinstance(value, Domain):
        from opendp.domains import domain_type
        rt_type = RuntimeType.parse(domain_type(value))

        if isinstance(rt_type, RuntimeType):
            if rt_type.origin == OptionDomain.__name__:
                value.__class__ = OptionDomain
            elif rt_type.origin == AtomDomain.__name__:
                value.__class__ = AtomDomain
            elif rt_type.origin == VectorDomain.__name__:
                value.__class__ = VectorDomain
        else:
            if rt_type == SeriesDomain.__name__:
                value.__class__ = SeriesDomain
            elif rt_type == LazyFrameDomain.__name__:
                value.__class__ = LazyFrameDomain
            elif rt_type == ExtrinsicDomain.__name__:
                value.__class__ = ExtrinsicDomain
        # if you fall through these cases, then it is just treated as a generic Domain

    if isinstance(value, Metric):
        from opendp.metrics import metric_type

        rt_type = RuntimeType.parse(metric_type(value))

        if isinstance(rt_type, RuntimeType):
            if rt_type.origin == FrameDistance.__name__:
                value.__class__ = FrameDistance
        else:
            if rt_type == SymmetricIdDistance.__name__:
                value.__class__ = SymmetricIdDistance
            elif rt_type == ChangeOneIdDistance.__name__:
                value.__class__ = ChangeOneIdDistance
            elif rt_type == ExtrinsicDistance.__name__:
                value.__class__ = ExtrinsicDistance
        # if you fall through these cases, then it is just treated as a generic Metric

    if isinstance(value, Measure):
        from opendp.measures import measure_type

        rt_type = RuntimeType.parse(measure_type(value))

        if isinstance(rt_type, RuntimeType):
            if rt_type.origin == "Approximate":
                value.__class__ = ApproximateDivergence
        # if you fall through these cases, then it is just treated as a generic Measure

    if isinstance(value, ctypes.c_void_p):
        # returned void pointers are interpreted as None
        return

    return value


def _slice_to_py(raw: FfiSlicePtr, type_name: Union[RuntimeType, str]) -> Any:
    """Convert from `raw` FfiSlicePtr to Python type.
    This is the postprocessing step after _object_to_slice that unloads data from a ctypes representation.
    External checks allow this function to assume that `raw` is compatible with the type_name type.

    :param raw: raw pointer to an FfiSlice that will be unloaded into a Python type
    :param type_name: Rust type name that determines the Python type to unload into
    :return: a standard Python reference-counted data type
    """

    if isinstance(type_name, str):
        if type_name in ATOM_MAP:
            return _slice_to_scalar(raw, type_name)

        if type_name == "BitVector":
            return _slice_to_bitvector(raw)

        if type_name == "ExtrinsicObject":
            return _slice_to_extrinsic(raw)

        if type_name == "String":
            return _slice_to_string(raw)

        if type_name == "LazyFrame":
            return _slice_to_lazyframe(raw)

        if type_name == "DataFrame":
            return _slice_to_dataframe(raw)

        if type_name == "Series":
            return _slice_to_series(raw)

        if type_name == "Expr":
            return _slice_to_expr(raw)

        if type_name == "ExprPlan":
            return _slice_to_exprplan(raw)

        if type_name == "Bound":
            return _slice_to_group_bound(raw)
        
        if type_name == "Margin":
            return _slice_to_margin(raw)
        
        if type_name == "Bounds":
            return _slice_to_vector(raw, RuntimeType("Vec", ["Bound"]))

        if type_name == "AnyObject":
            return _slice_to_anyobject(raw)

    if isinstance(type_name, RuntimeType):
        if type_name.origin == "Vec":
            return _slice_to_vector(raw, type_name)

        if type_name.origin == "Function":
            return _slice_to_function(raw)

        if type_name.origin == "HashMap":
            return _slice_to_hashmap(raw)

        if type_name.origin == "Tuple":
            return _slice_to_tuple(raw, type_name)

        if type_name.origin == "Option":
            return _slice_to_option(raw, type_name)

    raise UnknownTypeException(type_name)  # pragma: no cover


def _py_to_slice(value: Any, type_name: Union[RuntimeType, str]) -> FfiSlicePtr:
    """Convert from Python `value` to FfiSlicePtr.
    The initial preprocessing step for _slice_to_object that loads data into a ctypes representation.
    External checks allow this function to assume that `value` is compatible with the type_name type.

    :param value: data to load into an FfiSlice
    :param type_name: Rust type name to load value into.
    :return: pointer to an FfiSlice owned by Python.
    """
    if isinstance(type_name, str):
        if type_name in ATOM_MAP:
            return _scalar_to_slice(value, type_name)

        if type_name == "BitVector":
            return _bitvector_to_slice(value)

        if type_name == "ExtrinsicObject":
            return _extrinsic_to_slice(value)

        if type_name == "AnyMeasurement":
            return _wrap_in_slice(value, 1)

        if type_name == "String":
            return _string_to_slice(value)

        if type_name == "Series":
            return _series_to_slice(value)

        if type_name in {"LazyFrame", "DslPlan"}:
            return _lazyframe_to_slice(value)

        if type_name == "DataFrame":
            return _dataframe_to_slice(value)
        
        if type_name == "Margin":
            return _margin_to_slice(value)
            
        if type_name == "Expr":
            return _expr_to_slice(value)

        if type_name == "Bound":
            return _bound_to_slice(value)
        
        if type_name == "Bounds":
            return _vector_to_slice(value, RuntimeType("Vec", ["Bound"]))

    if isinstance(type_name, RuntimeType):
        if type_name.origin == "Vec":
            return _vector_to_slice(value, type_name)

        if type_name.origin == "HashMap":
            return _hashmap_to_slice(value, type_name)

        if type_name.origin == "Function":
            return _function_to_slice(value, type_name)

        if type_name.origin == "Tuple":
            return _tuple_to_slice(value, type_name)

    raise UnknownTypeException(type_name)  # pragma: no cover


def _scalar_to_slice(val, type_name: str) -> FfiSlicePtr:
    np = import_optional_dependency('numpy', raise_error=False)
    if np is not None and isinstance(val, np.ndarray):
        val = val.item()
    val = _check_and_cast_scalar(type_name, val)
    # ctypes.byref has edge-cases that cause use-after-free errors. ctypes.pointer fixes these edge-cases
    return _wrap_in_slice(ctypes.pointer(ATOM_MAP[type_name](val)), 1)


def _slice_to_scalar(raw: FfiSlicePtr, type_name: str):
    return ctypes.cast(raw.contents.ptr, ctypes.POINTER(ATOM_MAP[type_name])).contents.value # type: ignore[attr-defined]

def _extrinsic_to_slice(val) -> FfiSlicePtr:
    return _wrap_in_slice(ctypes.pointer(ExtrinsicObject(ctypes.py_object(val))), 1)

def _slice_to_extrinsic(raw: FfiSlicePtr):
    return ctypes.cast(raw.contents.ptr, ctypes.POINTER(ExtrinsicObject)).contents.ptr

def _string_to_slice(val: str) -> FfiSlicePtr:
    np = import_optional_dependency('numpy', raise_error=False)
    if np is not None and isinstance(val, np.ndarray):
        val = val.item()
    return _wrap_in_slice(ctypes.pointer(ctypes.c_char_p(val.encode())), 1)


def _slice_to_string(raw: FfiSlicePtr) -> str:
    value = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_char_p)).contents.value
    assert value is not None 
    return value.decode()


def _bitvector_to_slice(val: Sequence[Any]) -> FfiSlicePtr:
    np = import_optional_dependency('numpy', raise_error=False)
    if np is not None and isinstance(val, np.ndarray):
        val = val.tobytes()
    
    if not isinstance(val, (bytes, bytearray)):
        raise TypeError("Expected type is BitVector but input data is not bytes or bytearray.")  # pragma: no cover

    array = (ctypes.c_uint8 * len(val)).from_buffer_copy(val) # type: ignore[operator]
    return _wrap_in_slice(array, len(val) * 8)


def _slice_to_bitvector(raw: FfiSlicePtr) -> bytes:
    # raw.contents.len is the number of valid bits.
    # Division by -8 is ceiling rather than floor: the number of bytes in the buffer
    n_bytes = -(raw.contents.len // -8)
    buffer = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_uint8))[0:n_bytes] # type: ignore
    return bytes(buffer)


def _vector_to_slice(val: Sequence[Any], type_name: RuntimeType) -> FfiSlicePtr:
    if type_name.origin != 'Vec' or len(type_name.args) != 1:
        raise ValueError("type_name must be a Vec<_>")  # pragma: no cover

    inner_type_name = type_name.args[0]
    # TODO: can we use underlying numpy buffers directly?
    if not isinstance(val, list):
        try:
            val = list(val)
        except TypeError:
            raise TypeError(f"Expected type is {type_name} but input data is not a list.")

    inner_type_name = type_name.args[0]

    if isinstance(inner_type_name, RuntimeType) or inner_type_name in {"Expr", "Bound"}:
        c_repr = [py_to_c(v, c_type=AnyObjectPtr, type_name=inner_type_name) for v in val]
        array = (AnyObjectPtr * len(val))(*c_repr) # type: ignore[operator]
        ffislice = _wrap_in_slice(array, len(val))
        ffislice.depends_on(*c_repr)
        return ffislice

    if inner_type_name == "ExtrinsicObject":
        c_repr = [ExtrinsicObject(ctypes.py_object(v)) for v in val]
        array = (ExtrinsicObject * len(val))(*c_repr)
        ffi_slice = _wrap_in_slice(array, len(val))
        ffi_slice.depends_on(c_repr)
        return ffi_slice

    if inner_type_name == "SeriesDomain":
        # define the ctype of an array of domains
        domain_array_type = (Domain * len(val)) # type: ignore[operator]
        # create an instance of a ctype array of domains
        array = domain_array_type(*val) # type: ignore[operator]
        return _wrap_in_slice(array, len(val))
    
    # remaining inner types should be atomic
    val = [_check_and_cast_scalar(inner_type_name, v) for v in val]

    if inner_type_name == "String":
        def str_to_slice(val):
            return ctypes.c_char_p(val.encode())
        array = (ctypes.c_char_p * len(val))(*map(str_to_slice, val))
        return _wrap_in_slice(array, len(val))

    if inner_type_name not in ATOM_MAP:
        raise TypeError(f"Members must be one of {tuple(ATOM_MAP.keys())}. Found {inner_type_name}.")  # pragma: no cover

    array = (ATOM_MAP[inner_type_name] * len(val))(*val)  # type: ignore[operator]
    return _wrap_in_slice(array, len(val))


def _slice_to_vector(raw: FfiSlicePtr, type_name: RuntimeType) -> Sequence[Any]:
    if type_name.origin != 'Vec' or len(type_name.args) != 1:
        raise ValueError("type_name must be a Vec<_>")  # pragma: no cover

    inner_type_name = type_name.args[0]

    if inner_type_name in {'AnyObject', 'Expr', 'Bound'}:
        from opendp._data import ffislice_of_anyobjectptrs
        raw = ffislice_of_anyobjectptrs(raw)
        array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(AnyObjectPtr))[0:raw.contents.len]
        res = list(map(c_to_py, array))
        # when the top-level AnyObject is freed, it recursively frees all anyobjects inside of it
        # adjust the type of constituent AnyObjects so that __delete__ is not called when they are dropped
        for elem in array:
            elem.__class__ = ctypes.POINTER(AnyObject)
        return res

    if inner_type_name == 'ExtrinsicObject':
        array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ExtrinsicObject))[0:raw.contents.len]
        return list(map(lambda v: v.ptr, array))

    if inner_type_name == 'String':
        array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_char_p))[0:raw.contents.len]
        return list(map(lambda v: v.decode(), array))

    assert isinstance(inner_type_name, str), f"inner type must be string, found {inner_type_name} of type {type(inner_type_name)}"
    return ctypes.cast(raw.contents.ptr, ctypes.POINTER(ATOM_MAP[inner_type_name]))[0:raw.contents.len]


def _tuple_to_slice(val: tuple[Any, ...], type_name: Union[RuntimeType, str]) -> FfiSlicePtr:
    type_name = cast(RuntimeType, type_name)
    inner_type_names = type_name.args
    if not isinstance(val, tuple):
        raise TypeError("Cannot coerce a non-tuple type to a tuple")  # pragma: no cover

    if len(inner_type_names) != len(val):
        raise TypeError("type_name members must have same length as tuple")  # pragma: no cover

    if inner_type_names == ['f64', 'ExtrinsicObject']:
        score_ptr = ctypes.pointer(ctypes.c_double(val[0]))
        ext_obj = ctypes.pointer(ExtrinsicObject(ctypes.py_object(val[1])))

        cand_ptr = py_to_c(ext_obj, c_type=AnyObjectPtr, type_name="ExtrinsicObject")
        array = (ctypes.c_void_p * 2)(
            ctypes.cast(score_ptr, ctypes.c_void_p), 
            ctypes.cast(cand_ptr, ctypes.c_void_p), 
        )
        return _wrap_in_slice(ctypes.pointer(array), 2)

    for t in inner_type_names:
        if t not in ATOM_MAP:
            raise TypeError(f"Tuple members must be one of {set(ATOM_MAP.keys())}. Got {t}")  # pragma: no cover

    # check that actual type can be represented by the inner_type_name
    val = tuple(
        _check_and_cast_scalar(inner_type_name, val[i])
        for i, inner_type_name in zip(range(len(val)), inner_type_names)
    )

    # ctypes.byref has edge-cases that cause use-after-free errors. ctypes.pointer fixes these edge-cases
    ptr_data = (
        ctypes.cast(ctypes.pointer(ATOM_MAP[name](v)), ctypes.c_void_p) # type: ignore[index]
        for v, name in zip(val, inner_type_names))

    array = (ctypes.c_void_p * len(val))(*ptr_data)
    return _wrap_in_slice(ctypes.pointer(array), len(val))


def _slice_to_tuple(raw: FfiSlicePtr, type_name: RuntimeType) -> tuple[Any, ...]:
    inner_type_names = type_name.args
    void_array_ptr = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_void_p))
    ptr_data: list[ctypes.c_void_p] = void_array_ptr[0:raw.contents.len]

    if inner_type_names == ['PrivacyProfile', 'f64']:
        curve = ctypes.cast(ptr_data[0], AnyObjectPtr)
        delta = ctypes.cast(ptr_data[1], ctypes.POINTER(ctypes.c_double))
        return PrivacyProfile(curve), delta.contents.value
    
    if inner_type_names == ['f64', 'AnyObject']:
        score = ctypes.cast(ptr_data[0], ctypes.POINTER(ctypes.c_double))
        candidate_obj = ctypes.cast(ptr_data[1], AnyObjectPtr)
        candidate = c_to_py(c_to_py(candidate_obj))
        candidate_obj.__class__ = ctypes.POINTER(AnyObject) # type: ignore[assignment]
        return score.contents.value, candidate

    # tuple of instances of Python types
    return tuple(ctypes.cast(void_p, ctypes.POINTER(ATOM_MAP[name])).contents.value # type: ignore[index,attr-defined]
                 for void_p, name in zip(ptr_data, inner_type_names))


def _slice_to_option(raw: FfiSlicePtr, type_name: RuntimeType) -> Optional[Any]:
    if raw.contents.len == 0:
        return None
    return _slice_to_py(raw, type_name.args[0])


def _hashmap_to_slice(val: MutableMapping, type_name: RuntimeType) -> FfiSlicePtr:
    key_type, val_type = type_name.args
    if not isinstance(val, MutableMapping):
        raise TypeError(f"Expected type is {type_name} but input data is not a dict.")  # pragma: no cover

    val = {
        _check_and_cast_scalar(key_type, k):
            _check_and_cast_scalar(val_type, v) if val_type != "ExtrinsicObject" else v
        for k, v in val.items()
    }
    
    keys: AnyObjectPtr = py_to_c(list(val.keys()), type_name=f"Vec<{key_type}>", c_type=AnyObjectPtr)
    vals: AnyObjectPtr = py_to_c(list(val.values()), type_name=f"Vec<{val_type}>", c_type=AnyObjectPtr)
    ffislice = _wrap_in_slice(ctypes.pointer((AnyObjectPtr * 2)(keys, vals)), 2) # type: ignore[operator]

    # The __del__ destructor on `keys` and `vals` is called and memory freed when their refcounts go to zero.
    # ffislice needs keys and vals to have a lifetime at least as long as itself,
    # so we can't allow their refcount to go to zero until ffislice is freed.
    ffislice.depends_on(keys, vals)
    return ffislice


def _slice_to_function(raw: FfiSlicePtr) -> Function:
    # for ε(α)-RDP curves
    function = ctypes.cast(raw.contents.ptr, ctypes.POINTER(AnyFunction)).contents
    # put the contents behind a new, python pointer
    return ctypes.cast(ctypes.pointer(function), Function)


def _function_to_slice(raw: Function, type_name: RuntimeType) -> FfiSlicePtr:
    if not isinstance(raw, Function):
        from opendp.core import new_function
        raw = new_function(raw, TO=type_name.args[1])
    return _wrap_in_slice(raw, 1)


def _slice_to_hashmap(raw: FfiSlicePtr) -> MutableMapping:
    slice_array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(AnyObjectPtr))
    keys: AnyObjectPtr = slice_array[0]
    vals: AnyObjectPtr = slice_array[1]
    result = dict(zip(c_to_py(keys), c_to_py(vals)))

    # AnyObjectPtr.__del__ would free the memory behind keys and vals when this stack frame is popped.
    # But that memory has a lifetime at least as long as raw, so it cannot be freed yet.
    # Adjust the class to avoid calling AnyObjectPtr.__del__, which would free the backing memory.
    keys.__class__ = ctypes.POINTER(AnyObject) # type: ignore[assignment]
    vals.__class__ = ctypes.POINTER(AnyObject) # type: ignore[assignment]
    return result


def _lazyframe_to_slice(val) -> FfiSlicePtr:
    pl = import_optional_dependency('polars')
    if not isinstance(val, pl.LazyFrame):
        raise ValueError("expected Polars LazyFrame")

    state = val.__getstate__()
    raw = _wrap_in_slice(state, len(state))
    raw.depends_on(state)
    return raw


def _slice_to_lazyframe(raw: FfiSlicePtr):
    pl = import_optional_dependency('polars')
    lf = pl.LazyFrame()
    slice_array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_uint8))
    lf.__setstate__(bytes(slice_array[0:raw.contents.len]))
    return lf


def _expr_to_slice(val) -> FfiSlicePtr:
    pl = import_optional_dependency('polars')
    if isinstance(val, str):
        val = pl.col(val)

    if not isinstance(val, pl.Expr):
        raise ValueError("expected Polars Expr")

    state = val.__getstate__()
    raw = _wrap_in_slice(state, len(state))
    raw.depends_on(state)
    return raw


def _slice_to_expr(raw: FfiSlicePtr):
    pl = import_optional_dependency('polars')
    expr = pl.all()
    slice_array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_uint8))
    expr.__setstate__(bytes(slice_array[0:raw.contents.len]))
    return expr


def _slice_to_exprplan(raw: FfiSlicePtr):
    void_array_ptr = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_void_p))
    ptr_data: list[ctypes.c_void_p] = void_array_ptr[0:raw.contents.len]

    plan = _slice_to_lazyframe(ctypes.cast(ptr_data[0], FfiSlicePtr))
    expr = _slice_to_expr(ctypes.cast(ptr_data[1], FfiSlicePtr))
    fill = _slice_to_expr(ctypes.cast(ptr_data[2], FfiSlicePtr)) if raw.contents.len == 3 else None

    from collections import namedtuple

    ExprPlan = namedtuple("ExprPlan", ["plan", "expr", "fill"])
    return ExprPlan(plan, expr, fill)

def _to_optional_u32_void_ptr(v):
    if v is None:
        return
    check_c_int_cast(v, "u32")
    return ctypes.cast(ctypes.pointer(ctypes.c_uint32(v)), ctypes.c_void_p)


def _from_optional_u32_void_ptr(ptr):
    u32_ptr = ctypes.cast(ptr, ctypes.POINTER(ctypes.c_uint32))
    return u32_ptr[0] if u32_ptr else None


def _slice_to_margin(raw: FfiSlicePtr):
    from opendp.extras.polars import Margin

    void_array_ptr = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_void_p))
    ptr_data: list[ctypes.c_void_p] = void_array_ptr[0: raw.contents.len]

    invariant = ctypes.cast(ptr_data[3], ctypes.c_char_p)
    return Margin(
        by=c_to_py(ctypes.cast(ptr_data[0], AnyObjectPtr)),
        max_length=_from_optional_u32_void_ptr(ptr_data[1]),
        max_groups=_from_optional_u32_void_ptr(ptr_data[2]),
        invariant=invariant.value.decode() if invariant.value else None, # type: ignore[arg-type]
    )

def _check_polars_by(by):
    if isinstance(by, str):
        raise ValueError(f"by ({by}) must be a sequence type; Did you mean [\"{by}\"]?")
    
    if not isinstance(by, Sequence):
        raise ValueError(f"by ({by}) must be a sequence type")


def _margin_to_slice(val) -> FfiSlicePtr:
    from opendp.extras.polars import Margin
    if not isinstance(val, Margin):
        raise ValueError(f"expected Polars Margin, got {val}") # pragma: no cover
    
    _check_polars_by(val.by)
    by = ctypes.cast(py_to_c(val.by, c_type=AnyObjectPtr, type_name="Vec<Expr>"), ctypes.c_void_p)

    max_length = _to_optional_u32_void_ptr(val.max_length)
    max_groups = _to_optional_u32_void_ptr(val.max_groups)

    def str_to_nullptr(s):
        return ctypes.cast(ctypes.c_char_p(s.encode()), ctypes.c_void_p)

    invariant = str_to_nullptr(val.invariant) if val.invariant else None

    array = (ctypes.c_void_p * 4)(by, max_length, max_groups, invariant)
    return _wrap_in_slice(ctypes.pointer(array), 4)

def _slice_to_group_bound(raw: FfiSlicePtr):
    from opendp.extras.polars import Bound

    void_array_ptr = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_void_p))
    ptr_data: list[ctypes.c_void_p] = void_array_ptr[0: raw.contents.len]

    return Bound(
        by=c_to_py(ctypes.cast(ptr_data[0], AnyObjectPtr)),
        per_group=_from_optional_u32_void_ptr(ptr_data[1]),
        num_groups=_from_optional_u32_void_ptr(ptr_data[2]),
    )


def _bound_to_slice(val) -> FfiSlicePtr:
    from opendp.extras.polars import Bound
    assert isinstance(val, Bound)

    _check_polars_by(val.by)
    by = ctypes.cast(py_to_c(val.by, c_type=AnyObjectPtr, type_name="Vec<Expr>"), ctypes.c_void_p)

    num_groups = _to_optional_u32_void_ptr(val.num_groups)
    per_group = _to_optional_u32_void_ptr(val.per_group)

    array = (ctypes.c_void_p * 3)(by, per_group, num_groups)
    return _wrap_in_slice(ctypes.pointer(array), 3)



def _dataframe_to_slice(val) -> FfiSlicePtr:
    pl = import_optional_dependency('polars')
    if not isinstance(val, pl.DataFrame):
        raise ValueError("expected Polars DataFrame")  # pragma: no cover

    slices = list(_series_to_slice(s) for s in val.get_columns())
    raw = _wrap_in_slice(ctypes.pointer((FfiSlicePtr * val.width)(*slices)), val.width)
    # extend the lifetime of each series' slice to that of the frame slice
    raw.depends_on(slices)
    return raw

def _slice_to_dataframe(raw: FfiSlicePtr):
    pl = import_optional_dependency('polars')
    slice_array = ctypes.cast(raw.contents.ptr, FfiSlicePtr)
    series = [_slice_to_series(FfiSlicePtr(ffislice)) for ffislice in slice_array[0:raw.contents.len]]
    return pl.DataFrame(series)


def _series_to_slice(val) -> FfiSlicePtr:
    from opendp._data import new_arrow_array, arrow_array_free

    pl = import_optional_dependency('polars')
    if not isinstance(val, pl.Series):
        raise ValueError("expected Polars Series")

    raw = new_arrow_array(val.name)
    slice_array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_void_p))
    array_ptr, schema_ptr = slice_array[0:2]

    # make the conversion through PyArrow's private API
    # this changes the pointer's memory and is thus unsafe. In particular, `_export_to_c` can go out of bounds
    # NOTE: consider changing to PyCapsule when available. https://github.com/pola-rs/polars/issues/12530
    val.to_arrow()._export_to_c(array_ptr, schema_ptr)

    # when freeing the slice, also free up the memory of what's left behind the slice
    class ArrowArrayFFIBuffer(object):
        def __init__(self, ptr) -> None:
            self.ptr = ptr

        def __del__(self) -> None:
            arrow_array_free(self.ptr)

    raw.depends_on(ArrowArrayFFIBuffer(raw.contents.ptr))
    return raw


def _slice_to_series(raw: FfiSlicePtr):
    pl = import_optional_dependency('polars')
    pyarrow = import_optional_dependency('pyarrow')
    slice_array = ctypes.cast(raw.contents.ptr, ctypes.POINTER(ctypes.c_void_p))
    array_ptr, schema_ptr, name_ptr = slice_array[0:3]

    arrow_array = pyarrow.Array._import_from_c(array_ptr, schema_ptr)
    series = pl.from_arrow(arrow_array)

    name_bytes = ctypes.cast(name_ptr, ctypes.c_char_p).value
    if name_bytes is not None:
        series = series.rename(name_bytes.decode())
    return series


def _slice_to_anyobject(raw: FfiSlicePtr):
    """Unpack an AnyObject from a slice.

    Used to return an Option<AnyObject> from Rust to Python.
    """

    obj = ctypes.cast(raw.contents.ptr, AnyObjectPtr)
    ret = c_to_py(obj)
    # don't free obj, because it is owned by Rust
    # c_to_py cares that the type is AnyObject, so this needs to happen after c_to_py
    obj.__class__ = ctypes.POINTER(AnyObject) # type: ignore[assignment]
    
    return ret


def _wrap_in_slice(ptr, len_: int) -> FfiSlicePtr:
    return FfiSlicePtr(FfiSlice(ctypes.cast(ptr, ctypes.c_void_p), len_))


def _wrap_py_func(func, TO):
    from opendp._convert import c_to_py, py_to_c

    def wrapper_func(c_arg):
        try:
            # 1. convert AnyObject to Python type
            py_arg = c_to_py(c_arg)
            # don't free c_arg, because it is owned by Rust
            c_arg.__class__ = ctypes.POINTER(AnyObject)

            # 2. invoke the user-supplied function
            py_out = func(py_arg)

            # 3. convert back to an AnyObject
            c_out = py_to_c(py_out, c_type=AnyObjectPtr, type_name=TO)
            # don't free c_out, because we are giving ownership to Rust
            c_out.__class__ = ctypes.POINTER(AnyObject)

            # 4. pack up into an FfiResult
            lib.ffiresult_ok.argtypes = [ctypes.c_void_p]
            lib.ffiresult_ok.restype = ctypes.c_void_p
            return lib.ffiresult_ok(ctypes.addressof(c_out.contents))

        except Exception:
            import traceback
            lib.ffiresult_err.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
            lib.ffiresult_err.restype = ctypes.c_void_p
            return lib.ffiresult_err(
                ctypes.c_char_p("Continued stack trace from Exception in user-defined function".encode()),
                ctypes.c_char_p(traceback.format_exc().encode()),
            )

    c_wrapper_func = CallbackFnValue(wrapper_func)
    lifeline = ExtrinsicObject(ctypes.py_object(c_wrapper_func))

    return ctypes.pointer(CallbackFn(c_wrapper_func, lifeline))


# The output type cannot be an `ctypes.POINTER(FfiResult)` due to:
#   https://bugs.python.org/issue5710#msg85731
#                                   (answer         , query       , is_internal  )
TransitionFnValue = ctypes.CFUNCTYPE(ctypes.c_void_p, AnyObjectPtr, ctypes.c_bool)

class TransitionFn(ctypes.Structure):
    _fields_ = [
        ("callback", TransitionFnValue),
        ("lifeline", ExtrinsicObject)
    ]

class TransitionFnPtr(ctypes.POINTER(TransitionFn)): # type: ignore[misc]
    _type_ = TransitionFn


def _wrap_py_transition(py_transition, A):
    from opendp._convert import c_to_py, py_to_c

    # the indicator that a query is internal is oftentimes not needed
    if len(signature(py_transition).parameters) == 1:
        py_transition_old = py_transition
        py_transition = lambda q, _=None: py_transition_old(q)

    def wrapper_func(c_query, c_is_internal: ctypes.c_bool):
        try:
            # 1. convert to Python type
            py_query = c_to_py(c_query)
            py_is_internal = c_is_internal
            # don't free c_arg, because it is owned by Rust
            c_query.__class__ = ctypes.POINTER(AnyObject)

            # 2. invoke the user-supplied function
            py_out = py_transition(py_query, py_is_internal)

            # 3. convert back to an AnyObject
            c_out = py_to_c(py_out, c_type=AnyObjectPtr, type_name=A)
            # don't free c_out, because we are giving ownership to Rust
            c_out.__class__ = ctypes.POINTER(AnyObject)

            # 4. pack up into an FfiResult
            lib.ffiresult_ok.argtypes = [ctypes.c_void_p]
            lib.ffiresult_ok.restype = ctypes.c_void_p
            return lib.ffiresult_ok(ctypes.addressof(c_out.contents))

        except Exception:
            import traceback
            lib.ffiresult_err.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
            lib.ffiresult_err.restype = ctypes.c_void_p
            return lib.ffiresult_err(
                ctypes.c_char_p("Continued stack trace from Exception in user-defined function".encode()),
                ctypes.c_char_p(traceback.format_exc().encode()),
            )

    c_wrapper_func = TransitionFnValue(wrapper_func)
    lifeline = ExtrinsicObject(ctypes.py_object(c_wrapper_func))

    return ctypes.pointer(TransitionFn(c_wrapper_func, lifeline))



# Auto-generated. Do not edit!
'''
TODO!
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401

__all__ = [
    "arrow_array_free",
    "bool_free",
    "erfc",
    "extrinsic_object_free",
    "ffislice_of_anyobjectptrs",
    "fill_bytes",
    "new_arrow_array",
    "object_as_slice",
    "object_free",
    "object_type",
    "onceframe_collect",
    "onceframe_lazy",
    "privacy_profile_delta",
    "privacy_profile_epsilon",
    "slice_as_object",
    "slice_free",
    "str_free"
]


def arrow_array_free(
    this
):
    r"""Internal function. Free the memory associated with `this`, a slice containing an Arrow array, schema, and name.

    .. end-markdown

    :param this: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_data__arrow_array_free
    lib_function.argtypes = [ctypes.c_void_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'arrow_array_free',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def bool_free(
    this
):
    r"""Internal function. Free the memory associated with `this`, a bool.
    Used to clean up after the relation check.

    .. end-markdown

    :param this: A pointer to the bool to free.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_data__bool_free
    lib_function.argtypes = [BoolPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'bool_free',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def erfc(
    value: float
) -> float:
    r"""Internal function. Compute erfc.

    Used to prove an upper bound on the error of erfc.

    .. end-markdown

    :param value: 
    :type value: float
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_value = py_to_c(value, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_data__erfc
    lib_function.argtypes = [ctypes.c_double]
    lib_function.restype = ctypes.c_double

    output = c_to_py(lib_function(c_value))
    try:
        output.__opendp_dict__ = {
            '__function__': 'erfc',
            '__module__': 'data',
            '__kwargs__': {
                'value': value
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def extrinsic_object_free(
    this
):
    r"""Internal function. Free the memory associated with `this`, a string.
    Used to clean up after the type getter functions.

    .. end-markdown

    :param this: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_data__extrinsic_object_free
    lib_function.argtypes = [ExtrinsicObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'extrinsic_object_free',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def ffislice_of_anyobjectptrs(
    raw
):
    r"""Internal function. Converts an FfiSlice of AnyObjects to an FfiSlice of AnyObjectPtrs.

    .. end-markdown

    :param raw: A pointer to the slice to free.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_raw = py_to_c(raw, c_type=FfiSlicePtr, type_name=None)

    # Call library function.
    lib_function = lib.opendp_data__ffislice_of_anyobjectptrs
    lib_function.argtypes = [FfiSlicePtr]
    lib_function.restype = FfiResult

    output = unwrap(lib_function(c_raw), FfiSlicePtr)
    try:
        output.__opendp_dict__ = {
            '__function__': 'ffislice_of_anyobjectptrs',
            '__module__': 'data',
            '__kwargs__': {
                'raw': raw
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def fill_bytes(
    ptr,
    len: int
) -> bool:
    r"""Internal function. Populate the buffer behind `ptr` with `len` random bytes
    sampled from a cryptographically secure RNG.

    .. end-markdown

    :param ptr: 
    :param len: 
    :type len: int
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_ptr = ptr
    c_len = py_to_c(len, c_type=ctypes.c_size_t, type_name="usize")

    # Call library function.
    lib_function = lib.opendp_data__fill_bytes
    lib_function.argtypes = [ctypes.POINTER(ctypes.c_uint8), ctypes.c_size_t]
    lib_function.restype = ctypes.c_bool

    output = c_to_py(lib_function(c_ptr, c_len))
    try:
        output.__opendp_dict__ = {
            '__function__': 'fill_bytes',
            '__module__': 'data',
            '__kwargs__': {
                'ptr': ptr, 'len': len
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def new_arrow_array(
    name: str
):
    r"""Allocate an empty ArrowArray and ArrowSchema that Rust owns the memory for.
    The ArrowArray and ArrowSchema are initialized empty, and are populated by the bindings language.

    .. end-markdown

    :param name: The name of the ArrowArray. A clone of this string owned by Rust will be returned in the slice.
    :type name: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_name = py_to_c(name, c_type=ctypes.c_char_p, type_name=None)

    # Call library function.
    lib_function = lib.opendp_data__new_arrow_array
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_name), FfiSlicePtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'new_arrow_array',
            '__module__': 'data',
            '__kwargs__': {
                'name': name
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def object_as_slice(
    obj
):
    r"""Internal function. Unload data from an AnyObject into an FfiSlicePtr.

    .. end-markdown

    :param obj: A pointer to the AnyObject to unpack.
    :return: An FfiSlice that contains the data in FfiObject, but in a format readable in bindings languages.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_obj = py_to_c(obj, c_type=AnyObjectPtr, type_name=None)

    # Call library function.
    lib_function = lib.opendp_data__object_as_slice
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = unwrap(lib_function(c_obj), FfiSlicePtr)
    try:
        output.__opendp_dict__ = {
            '__function__': 'object_as_slice',
            '__module__': 'data',
            '__kwargs__': {
                'obj': obj
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def object_free(
    this
):
    r"""Internal function. Free the memory associated with `this`, an AnyObject.

    .. end-markdown

    :param this: A pointer to the AnyObject to free.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_data__object_free
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'object_free',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def object_type(
    this
) -> str:
    r"""Internal function. Retrieve the type descriptor string of an AnyObject.

    .. end-markdown

    :param this: A pointer to the AnyObject.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=AnyObjectPtr, type_name=None)

    # Call library function.
    lib_function = lib.opendp_data__object_type
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'object_type',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def onceframe_collect(
    onceframe
):
    r"""Internal function. Collects a DataFrame from a OnceFrame, exhausting the OnceFrame.

    .. end-markdown

    :param onceframe: The queryable holding a LazyFrame.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_onceframe = py_to_c(onceframe, c_type=AnyObjectPtr, type_name="AnyQueryable")

    # Call library function.
    lib_function = lib.opendp_data__onceframe_collect
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_onceframe), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'onceframe_collect',
            '__module__': 'data',
            '__kwargs__': {
                'onceframe': onceframe
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def onceframe_lazy(
    onceframe
):
    r"""Internal function. Extracts a LazyFrame from a OnceFrame,
    circumventing protections against multiple evaluations.

    Each collection consumes the entire allocated privacy budget.
    To remain DP at the advertised privacy level, only collect the LazyFrame once.


    Required features: `honest-but-curious`

    **Why honest-but-curious?:**

    The privacy guarantees only apply if:

    1. The LazyFrame (compute plan) is only ever executed once.
    2. The analyst does not observe ordering of rows in the output.
       
    To ensure that row ordering is not observed:

    1. Do not extend the compute plan with order-sensitive computations.
    2. Shuffle the output once collected ([in Polars sample all, with shuffling enabled](https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.sample.html)).

    .. end-markdown

    :param onceframe: The queryable holding a LazyFrame.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_onceframe = py_to_c(onceframe, c_type=AnyObjectPtr, type_name="AnyQueryable")

    # Call library function.
    lib_function = lib.opendp_data__onceframe_lazy
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_onceframe), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'onceframe_lazy',
            '__module__': 'data',
            '__kwargs__': {
                'onceframe': onceframe
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def privacy_profile_delta(
    curve,
    epsilon: float
):
    r"""Internal function. Use a PrivacyProfile to find epsilon at a given `epsilon`.

    .. end-markdown

    :param curve: The PrivacyProfile.
    :param epsilon: What to fix epsilon to compute delta.
    :type epsilon: float
    :return: Delta at a given `epsilon`.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_curve = py_to_c(curve, c_type=AnyObjectPtr, type_name=None)
    c_epsilon = py_to_c(epsilon, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_data__privacy_profile_delta
    lib_function.argtypes = [AnyObjectPtr, ctypes.c_double]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_curve, c_epsilon), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'privacy_profile_delta',
            '__module__': 'data',
            '__kwargs__': {
                'curve': curve, 'epsilon': epsilon
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def privacy_profile_epsilon(
    profile,
    delta: float
):
    r"""Internal function. Use an PrivacyProfile to find epsilon at a given `delta`.

    .. end-markdown

    :param profile: The PrivacyProfile.
    :param delta: What to fix delta to compute epsilon.
    :type delta: float
    :return: Epsilon at a given `delta`.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_profile = py_to_c(profile, c_type=AnyObjectPtr, type_name=None)
    c_delta = py_to_c(delta, c_type=ctypes.c_double, type_name="f64")

    # Call library function.
    lib_function = lib.opendp_data__privacy_profile_epsilon
    lib_function.argtypes = [AnyObjectPtr, ctypes.c_double]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_profile, c_delta), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': 'privacy_profile_epsilon',
            '__module__': 'data',
            '__kwargs__': {
                'profile': profile, 'delta': delta
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def slice_as_object(
    raw: FfiSlicePtr,
    T: str
):
    r"""Internal function. Load data from a `slice` into an AnyObject

    .. end-markdown

    :param raw: A pointer to the slice with data.
    :type raw: FfiSlicePtr
    :param T: The type of the data in the slice.
    :type T: str
    :return: An AnyObject that contains the data in `slice`. The AnyObject also captures rust type information.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    T = parse_or_infer(T, raw) # type: ignore

    # Convert arguments to c types.
    c_raw = py_to_c(raw, c_type=FfiSlicePtr, type_name=T)
    c_T = py_to_c(T, c_type=ctypes.c_char_p, type_name=None)

    # Call library function.
    lib_function = lib.opendp_data__slice_as_object
    lib_function.argtypes = [FfiSlicePtr, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = unwrap(lib_function(c_raw, c_T), AnyObjectPtr)
    try:
        output.__opendp_dict__ = {
            '__function__': 'slice_as_object',
            '__module__': 'data',
            '__kwargs__': {
                'raw': raw, 'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def slice_free(
    this
):
    r"""Internal function. Free the memory associated with `this`, an FfiSlicePtr.
    Used to clean up after object_as_slice.
    Frees the slice, but not what the slice references!

    .. end-markdown

    :param this: A pointer to the FfiSlice to free.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_data__slice_free
    lib_function.argtypes = [FfiSlicePtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'slice_free',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def str_free(
    this
):
    r"""Internal function. Free the memory associated with `this`, a string.
    Used to clean up after the type getter functions.

    .. end-markdown

    :param this: A pointer to the string to free.
    :type this: str
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_data__str_free
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'str_free',
            '__module__': 'data',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output



# type: ignore
# analogous to impl MakeNoise<DI, MI, MO> for DiscreteLaplace in Rust
class DiscreteLaplace:
    def make_noise(self, input_space: tuple[DI, MI]) -> Measurement[DI, DI_Carrier, MI, MO]:
        # an equivalent random variable specific to the atom dtype
        rv_nature = DI_Atom.new_distribution(self.scale, self.k) # |\label{line:rv-nature}|
        # build a measurement sampling from this equivalent distribution
        return rv_nature.make_noise(input_space) # |\label{line:make-noise}|



# type: ignore
def make_select_private_candidate(
    measurement: Measurement, 
    stop_probability: float, 
    threshold: float,
) -> Measurement:
    if not 0 <= stop_probability < 1: # |\label{stop-prob}|
        raise "stop_probability must be in [0, 1)"

    if not threshold.is_finite():
        raise "threshold must be finite"
    
    scale = None
    if stop_probability > 0.0:
        ln_cp = (1.0).neg_inf_sub(stop_probability).inf_ln()
        scale = ln_cp.recip().neg().into_rational() # |\label{scale}|

    def function(arg):
        remaining_iterations = None
        if scale is not None:
            remaining_iterations = UBig.ONE + sample_geometric_exp_fast(scale) # |\label{sample-geometric}|
        
        while True:
            score, output = measurement(arg)

            if score >= threshold:
                return score, output

            if remaining_iterations is not None:
                remaining_iterations -= UBig.ONE
                if remaining_iterations == UBig.ZERO:
                    return None

    return Measurement(
        input_domain=measurement.input_domain,
        input_metric=measurement.input_metric,
        output_measure=measurement.output_measure,
        function=function,
        privacy_map=lambda d_in: measurement.map(d_in).inf_mul(2),
    )



# type: ignore
# analogous to impl NoisePrivacyMap<L2Distance<RBig>, ZeroConcentratedDivergence> for ZExpFamily<1> in Rust
class ZExpFamily2:
    def noise_privacy_map(
        self, _input_metric: L2Distance[RBig], _output_measure: ZeroConcentratedDivergence
    ) -> PrivacyMap[L2Distance[RBig], ZeroConcentratedDivergence]:
        scale = self.scale
        if scale < RBig.ZERO:  # |\label{line:neg-scale}|
            raise "scale must be non-negative"

        def privacy_map(d_in: RBig):
            if d_in < RBig.ZERO:  # |\label{line:neg-sens}|
                raise "sensitivity must be non-negative"

            if d_in.is_zero():  # |\label{line:zero-sens}|
                return 0.0

            if scale.is_zero():  # |\label{line:zero-scale}|
                return float("inf")

            return f64.inf_cast((d_in / scale).pow(2) / rbig(2))  # |\label{line:map}|

        return PrivacyMap.new_fallible(privacy_map)



# Auto-generated. Do not edit!
'''
The ``metrics`` module provides fuctions that measure the distance between two elements of a domain.
For more context, see :ref:`metrics in the User Guide <metrics-user-guide>`.

For convenience, all the functions of this module are also available from :py:mod:`opendp.prelude`.
We suggest importing under the conventional name ``dp``:

.. code:: pycon

    >>> import opendp.prelude as dp
'''
from deprecated.sphinx import deprecated # noqa: F401 (Not every file actually has deprecated functions.)

from opendp._convert import *
from opendp._lib import *
from opendp.mod import *
from opendp.typing import *
from opendp.typing import _substitute # noqa: F401

__all__ = [
    "_change_one_id_distance_get_identifier",
    "_extrinsic_metric_descriptor",
    "_frame_distance_get_inner_metric",
    "_get_bound",
    "_metric_equal",
    "_metric_free",
    "_symmetric_id_distance_get_identifier",
    "absolute_distance",
    "change_one_distance",
    "change_one_id_distance",
    "discrete_distance",
    "frame_distance",
    "hamming_distance",
    "insert_delete_distance",
    "l01inf_distance",
    "l02inf_distance",
    "l1_distance",
    "l2_distance",
    "linf_distance",
    "metric_debug",
    "metric_distance_type",
    "metric_type",
    "symmetric_distance",
    "symmetric_id_distance",
    "user_distance"
]


def _change_one_id_distance_get_identifier(
    metric: Metric
):
    r"""Retrieve the identifier of a `ChangeOneIdDistance` metric.

    .. end-markdown

    :param metric: 
    :type metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_metric = py_to_c(metric, c_type=Metric, type_name="AnyMetric")

    # Call library function.
    lib_function = lib.opendp_metrics___change_one_id_distance_get_identifier
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_metric), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_change_one_id_distance_get_identifier',
            '__module__': 'metrics',
            '__kwargs__': {
                'metric': metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _extrinsic_metric_descriptor(
    metric: Metric
):
    r"""Retrieve the descriptor value stored in an extrinsic metric.

    .. end-markdown

    :param metric: The ExtrinsicDistance to extract the descriptor from
    :type metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_metric = py_to_c(metric, c_type=Metric, type_name="AnyMetric")

    # Call library function.
    lib_function = lib.opendp_metrics___extrinsic_metric_descriptor
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_metric), ExtrinsicObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_extrinsic_metric_descriptor',
            '__module__': 'metrics',
            '__kwargs__': {
                'metric': metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _frame_distance_get_inner_metric(
    metric: Metric
) -> Metric:
    r"""Retrieve the inner metric of a `FrameDistance` metric.

    .. end-markdown

    :param metric: 
    :type metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_metric = py_to_c(metric, c_type=Metric, type_name="AnyMetric")

    # Call library function.
    lib_function = lib.opendp_metrics___frame_distance_get_inner_metric
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_metric), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': '_frame_distance_get_inner_metric',
            '__module__': 'metrics',
            '__kwargs__': {
                'metric': metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _get_bound(
    bounds,
    by
):
    r"""Infer a bound when grouping by `by`.

    .. end-markdown

    :param bounds: 
    :param by: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_bounds = py_to_c(bounds, c_type=AnyObjectPtr, type_name="Bounds")
    c_by = py_to_c(by, c_type=AnyObjectPtr, type_name=RuntimeType(origin='Vec', args=["Expr"]))

    # Call library function.
    lib_function = lib.opendp_metrics___get_bound
    lib_function.argtypes = [AnyObjectPtr, AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_bounds, c_by), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_get_bound',
            '__module__': 'metrics',
            '__kwargs__': {
                'bounds': bounds, 'by': by
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _metric_equal(
    left: Metric,
    right: Metric
) -> bool:
    r"""Check whether two metrics are equal.

    .. end-markdown

    :param left: Metric to compare.
    :type left: Metric
    :param right: Metric to compare.
    :type right: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_left = py_to_c(left, c_type=Metric, type_name="AnyMetric")
    c_right = py_to_c(right, c_type=Metric, type_name="AnyMetric")

    # Call library function.
    lib_function = lib.opendp_metrics___metric_equal
    lib_function.argtypes = [Metric, Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_left, c_right), BoolPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_metric_equal',
            '__module__': 'metrics',
            '__kwargs__': {
                'left': left, 'right': right
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _metric_free(
    this
):
    r"""Internal function. Free the memory associated with `this`.

    .. end-markdown

    :param this: 
    :type this: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = this

    # Call library function.
    lib_function = lib.opendp_metrics___metric_free
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_void_p))
    try:
        output.__opendp_dict__ = {
            '__function__': '_metric_free',
            '__module__': 'metrics',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def _symmetric_id_distance_get_identifier(
    metric: Metric
):
    r"""Retrieve the identifier of a `SymmetricIdDistance` metric.

    .. end-markdown

    :param metric: 
    :type metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_metric = py_to_c(metric, c_type=Metric, type_name="AnyMetric")

    # Call library function.
    lib_function = lib.opendp_metrics___symmetric_id_distance_get_identifier
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_metric), AnyObjectPtr))
    try:
        output.__opendp_dict__ = {
            '__function__': '_symmetric_id_distance_get_identifier',
            '__module__': 'metrics',
            '__kwargs__': {
                'metric': metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def absolute_distance(
    T: RuntimeTypeDescriptor
) -> Metric:
    r"""Construct an instance of the `AbsoluteDistance` metric.

    [absolute_distance in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/metrics/fn.absolute_distance.html)

    .. end-markdown

    :param T: 
    :type T: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    T = RuntimeType.parse(type_name=T)

    # Convert arguments to c types.
    c_T = py_to_c(T, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_metrics__absolute_distance
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_T), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'absolute_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def change_one_distance(

) -> Metric:
    r"""Construct an instance of the `ChangeOneDistance` metric.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_metrics__change_one_distance
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'change_one_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def change_one_id_distance(
    identifier
) -> ChangeOneIdDistance:
    r"""Construct an instance of the `ChangeOneIdDistance` metric.

    .. end-markdown

    :param identifier: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_identifier = py_to_c(identifier, c_type=AnyObjectPtr, type_name="Expr")

    # Call library function.
    lib_function = lib.opendp_metrics__change_one_id_distance
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_identifier), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'change_one_id_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'identifier': identifier
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def discrete_distance(

) -> Metric:
    r"""Construct an instance of the `DiscreteDistance` metric.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_metrics__discrete_distance
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'discrete_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def frame_distance(
    inner_metric: Metric
) -> Metric:
    r"""`frame_distance` is a higher-order metric that contains multiple distance bounds for different groupings of data.

    .. end-markdown

    :param inner_metric: 
    :type inner_metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_inner_metric = py_to_c(inner_metric, c_type=Metric, type_name="AnyMetric")

    # Call library function.
    lib_function = lib.opendp_metrics__frame_distance
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_inner_metric), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'frame_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'inner_metric': inner_metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def hamming_distance(

) -> Metric:
    r"""Construct an instance of the `HammingDistance` metric.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_metrics__hamming_distance
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'hamming_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def insert_delete_distance(

) -> Metric:
    r"""Construct an instance of the `InsertDeleteDistance` metric.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_metrics__insert_delete_distance
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'insert_delete_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def l01inf_distance(
    metric: Metric
) -> Metric:
    r"""Construct an instance of the `L01InfDistance` metric.

    [l01inf_distance in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/metrics/fn.l01inf_distance.html)

    .. end-markdown

    :param metric: The metric used to compute distance between partitions.
    :type metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_metric = py_to_c(metric, c_type=Metric, type_name=None)

    # Call library function.
    lib_function = lib.opendp_metrics__l01inf_distance
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_metric), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'l01inf_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'metric': metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def l02inf_distance(
    metric: Metric
) -> Metric:
    r"""Construct an instance of the `L02InfDistance` metric.

    [l02inf_distance in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/metrics/fn.l02inf_distance.html)

    .. end-markdown

    :param metric: The metric used to compute distance between partitions.
    :type metric: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_metric = py_to_c(metric, c_type=Metric, type_name=None)

    # Call library function.
    lib_function = lib.opendp_metrics__l02inf_distance
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_metric), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'l02inf_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'metric': metric
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def l1_distance(
    T: RuntimeTypeDescriptor
) -> Metric:
    r"""Construct an instance of the `L1Distance` metric.

    [l1_distance in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/metrics/fn.l1_distance.html)

    .. end-markdown

    :param T: 
    :type T: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    T = RuntimeType.parse(type_name=T)

    # Convert arguments to c types.
    c_T = py_to_c(T, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_metrics__l1_distance
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_T), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'l1_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def l2_distance(
    T: RuntimeTypeDescriptor
) -> Metric:
    r"""Construct an instance of the `L2Distance` metric.

    [l2_distance in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/metrics/fn.l2_distance.html)

    .. end-markdown

    :param T: 
    :type T: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    T = RuntimeType.parse(type_name=T)

    # Convert arguments to c types.
    c_T = py_to_c(T, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_metrics__l2_distance
    lib_function.argtypes = [ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_T), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'l2_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def linf_distance(
    T: RuntimeTypeDescriptor,
    monotonic: bool = False
) -> Metric:
    r"""Construct an instance of the `LInfDistance` metric.

    [linf_distance in Rust documentation.](https://docs.rs/opendp/0.14.0/opendp/metrics/fn.linf_distance.html)

    .. end-markdown

    :param monotonic: set to true if non-monotonicity implies infinite distance
    :type monotonic: bool
    :param T: The type of the distance.
    :type T: :py:ref:`RuntimeTypeDescriptor`
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # Standardize type arguments.
    T = RuntimeType.parse(type_name=T)

    # Convert arguments to c types.
    c_monotonic = py_to_c(monotonic, c_type=ctypes.c_bool, type_name="bool")
    c_T = py_to_c(T, c_type=ctypes.c_char_p)

    # Call library function.
    lib_function = lib.opendp_metrics__linf_distance
    lib_function.argtypes = [ctypes.c_bool, ctypes.c_char_p]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_monotonic, c_T), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'linf_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'monotonic': monotonic, 'T': T
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def metric_debug(
    this: Metric
) -> str:
    r"""Debug a `metric`.

    .. end-markdown

    :param this: The metric to debug (stringify).
    :type this: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Metric, type_name=None)

    # Call library function.
    lib_function = lib.opendp_metrics__metric_debug
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'metric_debug',
            '__module__': 'metrics',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def metric_distance_type(
    this: Metric
) -> str:
    r"""Get the distance type of a `metric`.

    .. end-markdown

    :param this: The metric to retrieve the distance type from.
    :type this: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Metric, type_name=None)

    # Call library function.
    lib_function = lib.opendp_metrics__metric_distance_type
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'metric_distance_type',
            '__module__': 'metrics',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def metric_type(
    this: Metric
) -> str:
    r"""Get the type of a `metric`.

    .. end-markdown

    :param this: The metric to retrieve the type from.
    :type this: Metric
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_this = py_to_c(this, c_type=Metric, type_name=None)

    # Call library function.
    lib_function = lib.opendp_metrics__metric_type
    lib_function.argtypes = [Metric]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_this), ctypes.c_char_p))
    try:
        output.__opendp_dict__ = {
            '__function__': 'metric_type',
            '__module__': 'metrics',
            '__kwargs__': {
                'this': this
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def symmetric_distance(

) -> Metric:
    r"""Construct an instance of the `SymmetricDistance` metric.

    .. end-markdown


    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # No arguments to convert to c types.
    # Call library function.
    lib_function = lib.opendp_metrics__symmetric_distance
    lib_function.argtypes = []
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'symmetric_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def symmetric_id_distance(
    identifier
) -> SymmetricIdDistance:
    r"""Construct an instance of the `SymmetricIdDistance` metric.

    .. end-markdown

    :param identifier: 
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    # No type arguments to standardize.
    # Convert arguments to c types.
    c_identifier = py_to_c(identifier, c_type=AnyObjectPtr, type_name="Expr")

    # Call library function.
    lib_function = lib.opendp_metrics__symmetric_id_distance
    lib_function.argtypes = [AnyObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_identifier), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'symmetric_id_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'identifier': identifier
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output


def user_distance(
    identifier: str,
    descriptor = None
) -> Metric:
    r"""Construct a new UserDistance.
    Any two instances of an UserDistance are equal if their string descriptors are equal.


    Required features: `honest-but-curious`

    **Why honest-but-curious?:**

    Your definition of `d` must satisfy the requirements of a pseudo-metric:

    1. for any $x$, $d(x, x) = 0$
    2. for any $x, y$, $d(x, y) \ge 0$ (non-negativity)
    3. for any $x, y$, $d(x, y) = d(y, x)$ (symmetry)
    4. for any $x, y, z$, $d(x, z) \le d(x, y) + d(y, z)$ (triangle inequality)

    .. end-markdown

    :param identifier: A string description of the metric.
    :type identifier: str
    :param descriptor: Additional constraints on the domain.
    :raises TypeError: if an argument's type differs from the expected type
    :raises UnknownTypeException: if a type argument fails to parse
    :raises OpenDPException: packaged error from the core OpenDP library
    """
    assert_features("honest-but-curious")

    # No type arguments to standardize.
    # Convert arguments to c types.
    c_identifier = py_to_c(identifier, c_type=ctypes.c_char_p, type_name=None)
    c_descriptor = py_to_c(descriptor, c_type=ExtrinsicObjectPtr, type_name="ExtrinsicObject")

    # Call library function.
    lib_function = lib.opendp_metrics__user_distance
    lib_function.argtypes = [ctypes.c_char_p, ExtrinsicObjectPtr]
    lib_function.restype = FfiResult

    output = c_to_py(unwrap(lib_function(c_identifier, c_descriptor), Metric))
    try:
        output.__opendp_dict__ = {
            '__function__': 'user_distance',
            '__module__': 'metrics',
            '__kwargs__': {
                'identifier': identifier, 'descriptor': descriptor
            },
        }
    except AttributeError:  # pragma: no cover
        pass
    return output



# type: ignore
class FloatExpFamily:
    def make_noise(self, input_space) -> Measurement[AtomDomain[T], T, AbsoluteDistance[QI], MO]:
        scale, k = self.scale, self.k
        distribution = ZExpFamily(scale=integerize_scale(scale, k)) # |\label{line:dist}|

        t_int = make_float_to_bigint(input_space)
        m_noise = distribution.make_noise(t_int.output_space())
        return t_int >> m_noise >> then_deintegerize_vec(k)
    


# type: ignore
def sample_geometric_exp_fast(x) -> int:
    if x == 0:
        return 0
    
    s, t = Rational.into_numer_denom(x)
    
    while True:
        u = Integer.sample_uniform_int_below(t) # |\label{line:U}|
        d = bool.sample_bernoulli_exp(Rational(u, t)) # |\label{line:D}|
        if d:
            break
        
    v = sample_geometric_exp_slow(1) # |\label{line:V}|
    z = u + t * v
    return z // s


# type: ignore
# analogous to impl MakeNoiseThreshold<DI, MI, MO> for DiscreteGaussian in Rust
class DiscreteGaussian:
    def make_noise_threshold(self, input_space, threshold) -> Measurement[DI, DI_Carrier, MI, MO]:
        # an equivalent random variable specific to the atom dtype
        rv_nature = DI_Atom.new_distribution(self.scale, self.k) # |\label{line:rv-nature}|
        # build a measurement sampling from this equivalent distribution
        return rv_nature.make_noise_threshold(input_space, threshold) # |\label{line:make-noise}|



from opendp.mod import ApproximateDivergence
import opendp.prelude as dp


def test_approx_divergence():
    measure = dp.approximate(dp.max_divergence())
    assert isinstance(measure, ApproximateDivergence)
    assert measure.inner_measure == dp.max_divergence()


def test_inequality():
    measure = dp.max_divergence()
    assert measure != str(measure)



from opendp.extras._utilities import to_then
from opendp.mod import Domain, Metric, Transformation
from opendp.context import register
from opendp._lib import import_optional_dependency
from opendp._internal import _make_transformation


def make_np_clamp(
    input_domain: Domain, input_metric: Metric, norm, p, origin=None
) -> Transformation:
    """Construct a Transformation that clamps the norm of input data.

    :param input_domain: instance of `array2_domain(...)`
    :param input_metric: instance of `symmetric_distance()`
    :param norm: clamp each row to this norm. Required if data is not already bounded
    :param p: designates L`p` norm
    :param origin: norm clamping is centered on this point. Defaults to zero
    """
    import opendp.prelude as dp

    np = import_optional_dependency("numpy")

    dp.assert_features("contrib")

    if not str(input_domain).startswith("NPArray2Domain"):
        raise ValueError("input_domain must be NPArray2Domain")  # pragma: no cover

    norm = float(norm)
    if norm <= 0.0:
        raise ValueError("norm must positive")  # pragma: no cover
    if np.isnan(norm):
        raise ValueError("norm must not be NaN")  # pragma: no cover
    if p not in {1, 2}:
        raise ValueError("order p must be 1 or 2")  # pragma: no cover

    if origin is None:
        origin = 0.0

    if not np.all(np.isfinite(origin)):
        raise ValueError("origin must be finite")  # pragma: no cover

    def get_norm(x):
        with np.errstate(over="ignore"):
            current_norm = np.linalg.norm(x, ord=p, axis=1, keepdims=True)
        return np.nan_to_num(current_norm)

    def _function(arg):
        with np.errstate(over="ignore"):
            # don't mutate the input array
            arg = arg - origin
        arg = np.nan_to_num(arg)

        # may have to run multiple times due to FP rounding
        current_norm = get_norm(arg)
        while current_norm.max() > norm:
            with np.errstate(under="ignore", over="ignore"):
                factor = current_norm / norm
            arg /= np.maximum(np.nan_to_num(factor), 1)
            current_norm = get_norm(arg)

        return arg + origin

    kwargs = input_domain.descriptor._asdict() | {
        "norm": norm,
        "p": p,
        "origin": origin,
        "nan": False,
    }
    return _make_transformation(
        input_domain,
        input_metric,
        dp.numpy.array2_domain(**kwargs),
        input_metric,
        _function,
        lambda d_in: d_in,
    )


# generate then variant of the constructor
# TODO: Show this in the API Reference?
register(make_np_clamp)
then_np_clamp = to_then(make_np_clamp)
